#ifdef __PORT__
#include <SDL.h>
#include "port.h"
#endif
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#ifndef __PORT__
#include <io.h>
#endif
#include "save.h"
#include "cpu_68k.h"
#include "cpu_sh2.h"
#include "sh2.h"
#include "z80.h"
#include "cd_aspi.h"
#include "gens.h"
#include "g_main.h"
#include "g_ddraw.h"
#include "g_dsound.h"
#include "g_input.h"
#include "movie.h"
#include "vdp_io.h"
#include "vdp_rend.h"
#include "vdp_32x.h"
#include "rom.h"
#include "mem_m68k.h"
#include "mem_s68k.h"
#include "mem_sh2.h"
#include "mem_z80.h"
#include "ym2612.h"
#include "psg.h"
#include "pcm.h"
#include "pwm.h"
#include "scrshot.h"
#include "ggenie.h"
#include "io.h"
#include "misc.h"
#include "cd_sys.h"
#include <readline/readline.h>
#ifdef WITH_GTK
#include "glade/support.h"
#endif


#ifdef __PORT__
enum {
DIK_UP = SDLK_UP, DIK_DOWN = SDLK_DOWN,
DIK_LEFT = SDLK_LEFT, DIK_RIGHT = SDLK_RIGHT,
DIK_RETURN = SDLK_RETURN, DIK_A = SDLK_a,
DIK_S = SDLK_s, DIK_D = SDLK_d,
DIK_RSHIFT = SDLK_RSHIFT, DIK_Z = SDLK_z,
DIK_X = SDLK_x, DIK_C = SDLK_c, DIK_Y = SDLK_y,
DIK_H = SDLK_h, DIK_G = SDLK_g, DIK_J = SDLK_j,
DIK_U = SDLK_u, DIK_K = SDLK_k, DIK_L = SDLK_l,
DIK_M = SDLK_m, DIK_T = SDLK_t, DIK_I = SDLK_i,
DIK_O = SDLK_o, DIK_P = SDLK_p
};
#endif

#ifdef __PORT__
#define DIR_SEPARATOR '/'
#else
#define DIR_SEPARATOR '\\'
#endif


int Current_State = 0;
char State_Dir[1024] = "";
char SRAM_Dir[1024] = "";
char BRAM_Dir[1024] = "";
unsigned char State_Buffer[MAX_STATE_FILE_LENGHT];;

/* Get input filename via file selector dialog.
 * This and Change_File_S are identical.  -IM
 */
int Change_File_S(char *Dest, char *Dir, char *Titre, char *Filter, char *Ext)
{
#ifdef __PORT__
#ifdef WITH_GTK
	GtkWidget* filesel;
	gint res;
	gchar* filename=NULL;

	filesel = create_file_selector_private(NULL/*Filter*/,0);	
	gtk_window_set_title(GTK_WINDOW(filesel), Titre);
	res = gtk_dialog_run(GTK_DIALOG(filesel));
	
	switch(res)
	{
		case GTK_RESPONSE_OK:
			filename = (gchar*)gtk_file_selection_get_filename(GTK_FILE_SELECTION(filesel));
			break;
		case GTK_RESPONSE_CANCEL:
			break;
	}

	gtk_widget_destroy(filesel);
	if (filename)
	{
		strcpy(Dest,filename);
		return 1;
	}
    //Modif IM - meaningful return val not needed for non-gui file selection 
#else
    return 1;
#endif
#else

	OPENFILENAME ofn;

	SetCurrentDirectory(Gens_Path);

	if (!strcmp(Dest, ""))
	{
		strcpy(Dest, "default.");
		strcat(Dest, Ext);
	}

	memset(&ofn, 0, sizeof(OPENFILENAME));

	ofn.lStructSize = sizeof(OPENFILENAME);
	ofn.hwndOwner = HWnd;
	ofn.hInstance = ghInstance;
	ofn.lpstrFile = Dest;
	ofn.nMaxFile = 2047;
	ofn.lpstrFilter = Filter;
	ofn.nFilterIndex = 1;
	ofn.lpstrInitialDir = Dir;
	ofn.lpstrTitle = Titre;
	ofn.lpstrDefExt = Ext;
	ofn.Flags = 0;

	if (GetOpenFileName(&ofn)) return 1;
#endif
	return 0;
}

//Modif IM - both do the same thing anyway
int Change_File_L(char *Dest, char *Dir, char *Titre, char *Filter, char *Ext)
{
    return Change_File_S(Dest,Dir,Titre,Filter,Ext);
}


int Change_Dir(char *Dest, char *Dir, char *Titre, char *Filter, char *Ext)
{
#ifdef __PORT__
#ifdef WITH_GTK
	GtkWidget* filesel;
	gint res;
	gchar* filename=NULL;
	int i;

	filesel = create_file_selector_private(NULL/*Filter*/,0);	
	res = gtk_dialog_run(GTK_DIALOG(filesel));
	
	switch(res)
	{
		case GTK_RESPONSE_OK:
			filename = (gchar*)gtk_file_selection_get_filename(GTK_FILE_SELECTION(filesel));
			break;
		case GTK_RESPONSE_CANCEL:
			break;
	}

	gtk_widget_destroy(filesel);
#endif
#else

	OPENFILENAME ofn;
	int i;

	SetCurrentDirectory(Gens_Path);

	strcpy(Dest, "default.");
	strcat(Dest, Ext);
	memset(&ofn, 0, sizeof(OPENFILENAME));

	ofn.lStructSize = sizeof(OPENFILENAME);
	ofn.hwndOwner = HWnd;
	ofn.hInstance = ghInstance;
	ofn.lpstrFile = Dest;
	ofn.nMaxFile = 2047;
	ofn.lpstrFilter = Filter;
	ofn.nFilterIndex = 1;
	ofn.lpstrInitialDir = Dir;
	ofn.lpstrTitle = Titre;
	ofn.lpstrDefExt = Ext;
	ofn.Flags = 0;

#endif
#ifdef __PORT__
#ifdef WITH_GTK
	if (filename)
		{
		i = strlen(Dest) - 1;
		while ((i > 0) && (Dest[i] != DIR_SEPARATOR)) i--;
		if (!i) return 0;
		Dest[++i] = 0;
		return 1;
	}
#endif
#else
	if (GetSaveFileName(&ofn))
	{
		i = strlen(Dest) - 1;
		while ((i > 0) && (Dest[i] != DIR_SEPARATOR)) i--;
		if (!i) return 0;
		Dest[++i] = 0;
		return 1;
	}
#endif
	return 0;
}


FILE *Get_State_File()
{
	char Name[2048];
	char Ext[5] = ".gsX";

	SetCurrentDirectory(Gens_Path);

	Ext[3] = '0' + Current_State;
	strcpy(Name, State_Dir);
	strcat(Name, Rom_Name);
	strcat(Name, Ext);

	return fopen(Name, "rb");
}


void Get_State_File_Name(char *name)
{
	char Ext[5] = ".gsX";

	SetCurrentDirectory(Gens_Path);

	Ext[3] = '0' + Current_State;
	strcpy(name, State_Dir);
	strcat(name, Rom_Name);
	strcat(name, Ext);
}



int Load_State(char *Name)
{
	FILE *f;
	unsigned char *buf;
	int len;

	if (Genesis_Started) len = GENESIS_STATE_FILE_LENGHT_EX;
	else if (SegaCD_Started) len = SEGACD_STATE_FILE_LENGHT;
	else if (_32X_Started) len = G32X_STATE_FILE_LENGHT;
	else return 0;

	buf = State_Buffer;

	if ((f = fopen(Name, "rb")) == NULL) return 0;

	memset(buf, 0, len);
	if (fread(buf, 1, len, f))
	{
        int switched = 0; //Modif N/IM - switched is for displaying "switched to playback" message
        //Modif N - allow switching to playback if loading a state while recording in read-only mode
        if(MainMovie.ReadOnly == 1 && MainMovie.MovieStatus==MOVIE_RECORDING) 
        {
            MainMovie.MovieStatus=MOVIE_PLAYING;
            switched = 1;
        }

		z80_Reset(&M_Z80);
//		main68k_reset();
//		YM2612ResetChip(0);
//		Reset_VDP();

        //IM - probably need to do something here to make SegaCD savestates work
		//if (SegaCD_Started)
		//if (_32X_Started)

		Import_Genesis(buf);
		buf += 0x22500;
		if (SegaCD_Started)
		{
//			Import_SegaCD(buf);
			buf += 0xE0000;
		}
		if (_32X_Started)
		{
			Import_32X(buf);
			buf += 0x82A00;
		}

		Flag_Clr_Scr = 1;
		CRam_Flag = 1;
		VRam_Flag = 1;

        //Modif N/IM
        if(MainMovie.MovieStatus==MOVIE_PLAYING && MainMovie.ReadOnly == 0)
        {
            if(AutoBackupEnabled)
            {
                strncpy(Str_Tmp,MainMovie.FileName,512);
                strcat(MainMovie.FileName,".gmv");
                MainMovie.FileName[strlen(MainMovie.FileName)-7]='b';
                MainMovie.FileName[strlen(MainMovie.FileName)-6]='a';
                MainMovie.FileName[strlen(MainMovie.FileName)-5]='k';
                BackupMovieFile(&MainMovie);
                strncpy(MainMovie.FileName,Str_Tmp,512);
            }
            MainMovie.MovieStatus=MOVIE_RECORDING;
            sprintf(Str_Tmp, "STATE %d LOADED : MOVIE RECORDING RESUMED", Current_State);
        }
        else
        {
            if(switched) //Modif N - "switched to playback" message
                sprintf(Str_Tmp, "STATE %d LOADED : SWITCHED TO PLAYBACK", Current_State);
            else
                sprintf(Str_Tmp, "STATE %d LOADED", Current_State);
        }
        Put_Info(Str_Tmp, 2000);
	}
    //Modif N/IM - bulletproof recording (loading)
    if((MainMovie.MovieStatus == MOVIE_RECORDING) && MainMovie.File)
    {
        int m = fgetc(f);
        if(m == 'M' && !feof(f) && !ferror(f))
        {
            int pos = ftell(MainMovie.File);
            fseek(MainMovie.File,64,SEEK_SET);

            char* bla = new char [FrameCount*3];
            if(FrameCount*3 == fread(bla, 1, FrameCount*3, f))
                fwrite(bla, 1, FrameCount*3, MainMovie.File);
            delete[] bla;

            fseek(MainMovie.File,pos,SEEK_SET);
        }
    }
    
	fclose(f);

	return 1;
}


int Save_State(char *Name)
{
	FILE *f;
	unsigned char *buf;
	int len;

	buf = State_Buffer;

	if ((f = fopen(Name, "wb")) == NULL) return 0;

	if (Genesis_Started) len = GENESIS_STATE_FILE_LENGHT_EX;
	else if (SegaCD_Started) len = SEGACD_STATE_FILE_LENGHT;
	else if (_32X_Started) len = G32X_STATE_FILE_LENGHT;

	memset(buf, 0, len);
	Export_Genesis(buf);
	buf += 0x22500;
	if (SegaCD_Started)
	{
//		Export_SegaCD(buf);
		buf += 0xE0000;
	}
	if (_32X_Started)
	{
		Export_32X(buf);
		buf += 0x82A00;
	}
	
	fwrite(State_Buffer, 1, len, f);
    
    //Modif N/IM - bulletproof re-recording (saving)
    if((MainMovie.MovieStatus == MOVIE_RECORDING || MainMovie.MovieStatus == MOVIE_PLAYING) && MainMovie.File)
    {
        fputc('M', f);
        int pos = ftell(MainMovie.File);
        fseek(MainMovie.File,64,SEEK_SET);

        char* bla = new char [FrameCount*3];
        if(FrameCount*3 == fread(bla, 1, FrameCount*3, MainMovie.File))
            fwrite(bla, 1, FrameCount*3, f);
        delete[] bla;

        fseek(MainMovie.File,pos,SEEK_SET);
    }
    else
    {
        fputc('\0', f);
    }
    
	fclose(f);

	sprintf(Str_Tmp, "STATE %d SAVED", Current_State);
	Put_Info(Str_Tmp, 2000);

	return 1;
}


/*

GST genecyst save file

Range        Size   Description
-----------  -----  -----------
00000-00002  3      "GST"
00006-00007  2      "\xE0\x40"
000FA-00112  24     VDP registers
00112-00191  128    Color RAM
00192-001E1  80     Vertical scroll RAM
001E4-003E3  512    YM2612 registers
00474-02473  8192   Z80 RAM
02478-12477  65536  68K RAM
12478-22477  65536  Video RAM

main 68000 registers
--------------------

00080-0009F : D0-D7
000A0-000BF : A0-A7
000C8 : PC
000D0 : SR
000D2 : USP
000D6 : SSP

Z80 registers
-------------

00404 : AF
00408 : BC
0040C : DE
00410 : HL
00414 : IX
00418 : IY
0041C : PC
00420 : SP
00424 : AF'
00428 : BC'
0042C : DE'
00430 : HL'

00434 : I
00435 : Unknow
00436 : IFF1 = IFF2
00437 : Unknow

The 'R' register is not supported.

Z80 State
---------

00438 : Z80 RESET
00439 : Z80 BUSREQ
0043A : Unknow
0043B : Unknow

0043C : Z80 BANK (DWORD)

Gens and Kega ADD
-----------------

00040 : last VDP Control data written (DWORD)
00044 : second write flag (1 for second write) 
00045 : DMA Fill flag (1 mean next data write will cause a DMA fill)
00048 : VDP write address (DWORD)

00050 : Version       (Genecyst=0 ; Kega=5 ; Gens=5)
00051 : Emulator ID   (Genecyst=0 ; Kega=0 ; Gens=1)
00052 : System ID     (Genesis=0 ; SegaCD=1 ; 32X=2 ; SegaCD32X=3)

00060-00070 : PSG registers (WORD).


SEGA CD
-------

+00000-00FFF : Gate array & sub 68K
+01000-80FFF : Prg RAM
+81000-C0FFF : Word RAM (2M mode arrangement)
+C1000-D0FFF : PCM RAM
+D1000-DFFFF : CDD & CDC data (16 kB cache include)

32X
---

main SH2
--------

+00000-00FFF : cache
+01000-011FF : IO registers
+01200-0123F : R0-R15
+01240 : SR
+01244 : GBR
+01248 : VBR
+0124C : MACL
+01250 : MACH
+01254 : PR
+01258 : PC
+0125C : State

sub SH2
-------

+01400-023FF : cache
+02400-025FF : IO registers
+02600-0263F : R0-R15
+02640 : SR
+02644 : GBR
+02648 : VBR
+0264C : MACL
+02650 : MACH
+02654 : PR
+02658 : PC
+0265C : State

others
------
// Fix 32X save state :
// enregistrer correctement les registres systmes ...

+02700 : ADEN bit (bit 0)
+02701 : FM bit (bit 7)
+02702 : Master SH2 INT mask register
+02703 : Slave SH2 INT mask register
+02704 : 68000 32X rom bank register
+02705 : RV (Rom to VRAM DMA allowed) bit (bit 0)
+02710-0273F : FIFO stuff (not yet done)
+02740-0274F : 32X communication buffer
+02750-02759 : PWM registers
+02760-0276F : 32X VDP registers
+02800-029FF : 32X palette
+02A00-429FF : SDRAM
+42A00-829FF : FB1 & FB2

*/

void Import_Genesis(unsigned char *Data)
{
	unsigned char Reg_1[0x200], Version, *src;
	int i;

	VDP_Int = 0;
	DMAT_Lenght = 0;
	Version = Data[0x50];

	for(i = 0; i < 0x80; i++) CRam[i] = Data[i + 0x112];
	for(i = 0; i < 0x50; i++) VSRam[i] = Data[i + 0x192];
	for(i = 0; i < 0x2000; i++) Ram_Z80[i] = Data[i + 0x474];
	
	for(i = 0; i < 0x10000; i += 2)
	{
		Ram_68k[i + 0] = Data[i + 0x2478 + 1];
		Ram_68k[i + 1] = Data[i + 0x2478 + 0];
	}

	for(i = 0; i < 0x10000; i += 2)
	{
		VRam[i + 0] = Data[i + 0x12478 + 1];
		VRam[i + 1] = Data[i + 0x12478 + 0];
	}

	for(i = 0; i < 0x200; i++) Reg_1[i] = Data[i + 0x1E4];
	YM2612_Restore(Reg_1);

	if ((Version >= 2) && (Version < 4))
	{
		src = (unsigned char *) &Ctrl;
		for(i = 0; i < 7 * 4; i++) *src++ = Data[i + 0x30];

		Z80_State &= ~6;
		if (Data[0x440] & 1) Z80_State |= 2;
		if (Data[0x444] & 1) Z80_State |= 4;

		src = (unsigned char *) &Bank_Z80;
		for(i = 0; i < 4; i++) *src++ = Data[i + 0x448];

		src = (unsigned char *) &PSG_Save;
		for(i = 0; i < 8 * 4; i++) *src++ = Data[i + 0x224B8];
		PSG_Restore_State();
	}
	else if ((Version >= 4) || (Version == 0)) 		// New version compatible with Kega.
	{
		Z80_State &= ~6;

		if (Version == 4)
		{
			M_Z80.IM = Data[0x437];
			M_Z80.IFF.b.IFF1 = (Data[0x438] & 1) << 2;
			M_Z80.IFF.b.IFF2 = (Data[0x438] & 1) << 2;

			Z80_State |= (Data[0x439] & 1) << 1;
		}
		else
		{
			M_Z80.IM = 1;
			M_Z80.IFF.b.IFF1 = (Data[0x436] & 1) << 2;
			M_Z80.IFF.b.IFF2 = (Data[0x436] & 1) << 2;

			Z80_State |= ((Data[0x439] & 1) ^ 1) << 1;
			Z80_State |= ((Data[0x438] & 1) ^ 1) << 2;
		}

		src = (unsigned char *) &Ctrl;
		for(i = 0; i < 7 * 4; i++) *src++ = 0;

		Write_VDP_Ctrl(Data[0x40] + (Data[0x41] << 8));
		Write_VDP_Ctrl(Data[0x42] + (Data[0x43] << 8));

		Ctrl.Flag = Data[0x44];
		Ctrl.DMA = (Data[0x45] & 1) << 2;
		Ctrl.Address = Data[0x48] + (Data[0x49] << 8);
		
		src = (unsigned char *) &Bank_Z80;
		for(i = 0; i < 4; i++) *src++ = Data[i + 0x43C];

		if (Version >= 4)
		{
			for(i = 0; i < 8; i++) PSG_Save[i] = Data[i * 2 + 0x60] + (Data[i * 2 + 0x61] << 8);
			PSG_Restore_State();
		}
	}

	z80_Set_AF(&M_Z80, Data[0x404] + (Data[0x405] << 8));
	M_Z80.BC.w.BC = Data[0x408] + (Data[0x409] << 8);
	M_Z80.DE.w.DE = Data[0x40C] + (Data[0x40D] << 8);
	M_Z80.HL.w.HL = Data[0x410] + (Data[0x411] << 8);
	M_Z80.IX.w.IX = Data[0x414] + (Data[0x415] << 8);
	M_Z80.IY.w.IY = Data[0x418] + (Data[0x419] << 8);
	z80_Set_PC(&M_Z80, Data[0x41C] + (Data[0x41D] << 8));
	M_Z80.SP.w.SP = Data[0x420] + (Data[0x421] << 8);
	z80_Set_AF2(&M_Z80, Data[0x424] + (Data[0x425] << 8));
	M_Z80.BC2.w.BC2 = Data[0x428] + (Data[0x429] << 8);
	M_Z80.DE2.w.DE2 = Data[0x42C] + (Data[0x42D] << 8);
	M_Z80.HL2.w.HL2 = Data[0x430] + (Data[0x431] << 8);
	M_Z80.I = Data[0x434] & 0xFF;

    //Modif IM/N - update frame count when loading a state
    FrameCount=Data[0x22478]+(Data[0x22479]<<8)+(Data[0x2247A]<<16)+(Data[0x2247B]<<24);
    MainMovie.NbRerecords++;
    if(MainMovie.MovieStatus==MOVIE_RECORDING)
        MainMovie.LastFrame=FrameCount;
    
	main68k_GetContext(&Context_68K);

	for(i = 0; i < 24; i++) Set_VDP_Reg(i, Data[0xFA + i]);
	
	src = (unsigned char *) &Context_68K.dreg[0];
	for(i = 0; i < 8 * 2 * 4; i++) *src++ = Data[0x80 + i];

	src = (unsigned char *) &Context_68K.pc;
	for(i = 0; i < 4; i++) *src++ = Data[0xC8 + i];

	src = (unsigned char *) &Context_68K.sr;
	for(i = 0; i < 2; i++) *src++ = Data[0xD0 + i];

	if ((Version >= 3) || (Version == 0))
	{
		if (Data[0xD1] & 0x20)
		{
			// Supervisor
			src = (unsigned char *) &Context_68K.asp;
			for(i = 0; i < 2; i++) *src++ = Data[0xD2 + i];
		}
		else
		{
			// User
			src = (unsigned char *) &Context_68K.asp;
			for(i = 0; i < 2; i++) *src++ = Data[0xD6 + i];
		}
	}

	main68k_SetContext(&Context_68K);
}


void Export_Genesis(unsigned char *Data)
{
	unsigned char Reg_1[0x200], *src;
	int i;

	while (DMAT_Lenght) Update_DMA();		// Be sure to finish DMA before save

	Data[0x00] = 'G';
	Data[0x01] = 'S';
	Data[0x02] = 'T';
	Data[0x03] = 0x40;
	Data[0x04] = 0xE0;

	Data[0x50] = 5;		// Version
	Data[0x51] = 0;		// Gens

	PSG_Save_State();

	for(i = 0; i < 8; i++)
	{
		Data[0x60 + i * 2] = PSG_Save[i] & 0xFF;
		Data[0x61 + i * 2] = (PSG_Save[i] >> 8) & 0xFF;
	}

	main68k_GetContext(&Context_68K);

	src = (unsigned char *) &(Context_68K.dreg[0]);
	for(i = 0; i < 8 * 2 * 4; i++) Data[0x80 + i] = *src++;

	src = (unsigned char *) &(Context_68K.pc);
	for(i = 0; i < 4; i++) Data[0xC8 + i] = *src++;

	src = (unsigned char *) &(Context_68K.sr);
	for(i = 0; i < 2; i++) Data[0xD0 + i] = *src++;

	if (Context_68K.sr & 0x2000)
	{
		src = (unsigned char *) &(Context_68K.asp);
		for(i = 0; i < 4; i++) Data[0xD2 + i] = *src++;
		src = (unsigned char *) &(Context_68K.areg[7]);
		for(i = 0; i < 4; i++) Data[0xD6 + i] = *src++;
	}
	else
	{
		src = (unsigned char *) &(Context_68K.asp);
		for(i = 0; i < 4; i++) Data[0xD6 + i] = *src++;
		src = (unsigned char *) &(Context_68K.areg[7]);
		for(i = 0; i < 4; i++) Data[0xD2 + i] = *src++;
	}

	src = (unsigned char *) &(Ctrl.Data);
	for(i = 0; i < 4; i++) Data[0x40 + i] = *src++;

	Data[0x44] = Ctrl.Flag;
	Data[0x45] = (Ctrl.DMA >> 2) & 1;

	Data[0x48] = Ctrl.Address & 0xFF;
	Data[0x49] = (Ctrl.Address >> 8) & 0xFF;

	VDP_Reg.DMA_Lenght_L = VDP_Reg.DMA_Lenght & 0xFF;
	VDP_Reg.DMA_Lenght_H = (VDP_Reg.DMA_Lenght >> 8) & 0xFF;

	VDP_Reg.DMA_Src_Adr_L = VDP_Reg.DMA_Address & 0xFF;
	VDP_Reg.DMA_Src_Adr_M = (VDP_Reg.DMA_Address >> 8) & 0xFF;
	VDP_Reg.DMA_Src_Adr_H = (VDP_Reg.DMA_Address >> 16) & 0xFF;

	VDP_Reg.DMA_Src_Adr_H |= Ctrl.DMA_Mode & 0xC0;

	src = (unsigned char *) &(VDP_Reg.Set1);
	for(i = 0; i < 24; i++)
	{
		Data[0xFA + i] = *src;
		src += 4;
	}

	for(i = 0; i < 0x80; i++) Data[i + 0x112] = CRam[i];
	for(i = 0; i < 0x50; i++) Data[i + 0x192] = VSRam[i];

	YM2612_Save(Reg_1);
	for(i = 0; i < 0x200; i++) Data[i + 0x1E4] = Reg_1[i];

	Data[0x404] = (unsigned char) (z80_Get_AF(&M_Z80) & 0xFF);
	Data[0x405] = (unsigned char) (z80_Get_AF(&M_Z80) >> 8);
	Data[0x408] = (unsigned char) (M_Z80.BC.w.BC & 0xFF);
	Data[0x409] = (unsigned char) (M_Z80.BC.w.BC >> 8);
	Data[0x40C] = (unsigned char) (M_Z80.DE.w.DE & 0xFF);
	Data[0x40D] = (unsigned char) (M_Z80.DE.w.DE >> 8);
	Data[0x410] = (unsigned char) (M_Z80.HL.w.HL & 0xFF);
	Data[0x411] = (unsigned char) (M_Z80.HL.w.HL >> 8);
	Data[0x414] = (unsigned char) (M_Z80.IX.w.IX & 0xFF);
	Data[0x415] = (unsigned char) (M_Z80.IX.w.IX >> 8);
	Data[0x418] = (unsigned char) (M_Z80.IY.w.IY & 0xFF);
	Data[0x419] = (unsigned char) (M_Z80.IY.w.IY >> 8);
	Data[0x41C] = (unsigned char) (z80_Get_PC(&M_Z80) & 0xFF);
	Data[0x41D] = (unsigned char) (z80_Get_PC(&M_Z80) >> 8);
	Data[0x420] = (unsigned char) (M_Z80.SP.w.SP & 0xFF);
	Data[0x421] = (unsigned char) (M_Z80.SP.w.SP >> 8);
	Data[0x424] = (unsigned char) (z80_Get_AF2(&M_Z80) & 0xFF);
	Data[0x425] = (unsigned char) (z80_Get_AF2(&M_Z80) >> 8);
	Data[0x428] = (unsigned char) (M_Z80.BC2.w.BC2 & 0xFF);
	Data[0x429] = (unsigned char) (M_Z80.BC2.w.BC2 >> 8);
	Data[0x42C] = (unsigned char) (M_Z80.DE2.w.DE2 & 0xFF);
	Data[0x42D] = (unsigned char) (M_Z80.DE2.w.DE2 >> 8);
	Data[0x430] = (unsigned char) (M_Z80.HL2.w.HL2 & 0xFF);
	Data[0x431] = (unsigned char) (M_Z80.HL2.w.HL2 >> 8);
	Data[0x434] = (unsigned char) (M_Z80.I);
	Data[0x436] = (unsigned char) (M_Z80.IFF.b.IFF1 >> 2);

	Data[0x438] = (unsigned char) (((Z80_State & 4) >> 2) ^ 1);
	Data[0x439] = (unsigned char) (((Z80_State & 2) >> 1) ^ 1);

	src = (unsigned char *) &Bank_Z80;
	for(i = 0; i < 4; i++) Data[0x43C + i] = *src++;

	for(i = 0; i < 0x2000; i++) Data[i + 0x474] = Ram_Z80[i];

	for(i = 0; i < 0x10000; i += 2)
	{
		Data[i + 0x2478 + 1] = Ram_68k[i + 0];
		Data[i + 0x2478 + 0] = Ram_68k[i + 1];
	}

	for(i = 0; i < 0x10000; i += 2)
	{
		Data[i + 0x12478 + 1] = VRam[i + 0];
		Data[i + 0x12478 + 0] = VRam[i + 1];
	}
    //Modif N/IM - update frame count when loading
    Data[0x22478]=(FrameCount&0xFF);   
    Data[0x22479]=(FrameCount>>8)&0xFF; 
    Data[0x2247A]=(FrameCount>>16)&0xFF;
    Data[0x2247B]=(FrameCount>>24)&0xFF;
}


void Import_32X(unsigned char *Data)
{
	unsigned char *src;
	int i;

	for(i = 0; i < 0x1000; i++) M_SH2.Cache[i] = Data[i];
	for(i = 0; i < 0x200; i++) SH2_Write_Byte(&M_SH2, 0xFFFFFE00 + i, Data[0x1000 + i]);

	src = (unsigned char *) &(M_SH2.R[0]);
	for(i = 0; i < 16 * 4; i++) *src++ = Data[0x1200 + i];

	SH2_Set_SR(&M_SH2, (Data[0x1243] << 24) + (Data[0x1242] << 16) + (Data[0x1241] << 8) + Data[0x1240]);
	SH2_Set_GBR(&M_SH2, (Data[0x1247] << 24) + (Data[0x1246] << 16) + (Data[0x1245] << 8) + Data[0x1244]);
	SH2_Set_VBR(&M_SH2, (Data[0x124B] << 24) + (Data[0x124A] << 16) + (Data[0x1249] << 8) + Data[0x1248]);
	SH2_Set_MACL(&M_SH2, (Data[0x124F] << 24) + (Data[0x124E] << 16) + (Data[0x124D] << 8) + Data[0x124C]);
	SH2_Set_MACH(&M_SH2, (Data[0x1253] << 24) + (Data[0x1252] << 16) + (Data[0x1251] << 8) + Data[0x1250]);
	SH2_Set_PR(&M_SH2, (Data[0x1257] << 24) + (Data[0x1256] << 16) + (Data[0x1255] << 8) + Data[0x1254]);
	SH2_Set_PC(&M_SH2, (Data[0x125B] << 24) + (Data[0x125A] << 16) + (Data[0x1259] << 8) + Data[0x1258]);
	M_SH2.SH2Status = (Data[0x125F] << 24) + (Data[0x125E] << 16) + (Data[0x125D] << 8) + Data[0x125C];

	for(i = 0; i < 0x1000; i++) S_SH2.Cache[i] = Data[i + 0x1400];
	for(i = 0; i < 0x200; i++) SH2_Write_Byte(&S_SH2, 0xFFFFFE00 + i, Data[0x2400 + i]);

	src = (unsigned char *) &(S_SH2.R[0]);
	for(i = 0; i < 16 * 4; i++) *src++ = Data[0x2600 + i];

	SH2_Set_SR(&S_SH2, (Data[0x2643] << 24) + (Data[0x2642] << 16) + (Data[0x2641] << 8) + Data[0x2640]);
	SH2_Set_GBR(&S_SH2, (Data[0x2647] << 24) + (Data[0x2646] << 16) + (Data[0x2645] << 8) + Data[0x2644]);
	SH2_Set_VBR(&S_SH2, (Data[0x264B] << 24) + (Data[0x264A] << 16) + (Data[0x2649] << 8) + Data[0x2648]);
	SH2_Set_MACL(&S_SH2, (Data[0x264F] << 24) + (Data[0x264E] << 16) + (Data[0x264D] << 8) + Data[0x264C]);
	SH2_Set_MACH(&S_SH2, (Data[0x2653] << 24) + (Data[0x2652] << 16) + (Data[0x2651] << 8) + Data[0x2650]);
	SH2_Set_PR(&S_SH2, (Data[0x2657] << 24) + (Data[0x2656] << 16) + (Data[0x2655] << 8) + Data[0x2654]);
	SH2_Set_PC(&S_SH2, (Data[0x265B] << 24) + (Data[0x265A] << 16) + (Data[0x2659] << 8) + Data[0x2658]);
	S_SH2.SH2Status = (Data[0x265F] << 24) + (Data[0x265E] << 16) + (Data[0x265D] << 8) + Data[0x265C];

	_32X_ADEN = Data[0x2700] & 1;
	_32X_RV = Data[0x2705] & 1;
	M68K_32X_Mode();
	_32X_FM = Data[0x2701] & 0x80;
	_32X_Set_FB();
	_32X_MINT = Data[0x2702];
	_32X_SINT = Data[0x2703];
	Bank_SH2 = Data[0x2704];
	M68K_Set_32X_Rom_Bank();
	
	/*******
	FIFO stuff to add here...	
	 *******/
	for(i = 0; i < 0x10; i++) SH2_Write_Byte(&M_SH2, 0x4020 + i, Data[0x2740 + i]);
	for(i = 0; i < 4; i++) SH2_Write_Byte(&M_SH2, 0x4030 + i, Data[0x2750 + i]);
	/*******
	Extra PWM stuff to add here...	
	 *******/

	// Do it to allow VDP write on 32X side
	_32X_FM = 0x80;
	_32X_Set_FB();

	for(i = 0; i < 0x10; i++) SH2_Write_Byte(&M_SH2, 0x4100 + i, Data[0x2760 + i]);
	for(i = 0; i < 0x200; i += 2) SH2_Write_Word(&M_SH2, 0x4200 + i, Data[0x2800 + i + 0] + (Data[0x2800 + i + 1] << 8));
	for(i = 0; i < 0x40000; i++) SH2_Write_Byte(&M_SH2, 0x6000000 + i, Data[0x2A00 + i]);

	_32X_FM = Data[0x2701] & 0x80;
	_32X_Set_FB();

	if (SH2_Read_Word(&M_SH2, 0x410A) & 1)
	{
		for(i = 0; i < 0x20000; i++) _32X_VDP_Ram[i] = Data[0x42A00 + i];
		for(i = 0; i < 0x20000; i++) _32X_VDP_Ram[i + 0x20000] = Data[0x62A00 + i];
	}
	else
	{
		for(i = 0; i < 0x20000; i++) _32X_VDP_Ram[i + 0x20000] = Data[0x42A00 + i];
		for(i = 0; i < 0x20000; i++) _32X_VDP_Ram[i] = Data[0x62A00 + i];
	}
}


void Export_32X(unsigned char *Data)
{
	unsigned char *src;
	int i;

	for(i = 0; i < 0x1000; i++) Data[i] = M_SH2.Cache[i];
	for(i = 0; i < 0x200; i++) Data[0x1000 + i] = M_SH2.IO_Reg[i];

	src = (unsigned char *) &(M_SH2.R[0]);
	for(i = 0; i < 16 * 4; i++) Data[0x1200 + i] = *src++;

	i = SH2_Get_SR(&M_SH2);
	Data[0x1240] = (unsigned char) (i & 0xFF);
	Data[0x1241] = (unsigned char) ((i >> 8) & 0xFF);
	Data[0x1242] = (unsigned char) ((i >> 16) & 0xFF);
	Data[0x1243] = (unsigned char) ((i >> 24) & 0xFF);
	i = SH2_Get_GBR(&M_SH2);
	Data[0x1244] = (unsigned char) (i & 0xFF);
	Data[0x1245] = (unsigned char) ((i >> 8) & 0xFF);
	Data[0x1246] = (unsigned char) ((i >> 16) & 0xFF);
	Data[0x1247] = (unsigned char) ((i >> 24) & 0xFF);
	i = SH2_Get_VBR(&M_SH2);
	Data[0x1248] = (unsigned char) (i & 0xFF);
	Data[0x1249] = (unsigned char) ((i >> 8) & 0xFF);
	Data[0x124A] = (unsigned char) ((i >> 16) & 0xFF);
	Data[0x124B] = (unsigned char) ((i >> 24) & 0xFF);
	i = SH2_Get_MACL(&M_SH2);
	Data[0x124C] = (unsigned char) (i & 0xFF);
	Data[0x124D] = (unsigned char) ((i >> 8) & 0xFF);
	Data[0x124E] = (unsigned char) ((i >> 16) & 0xFF);
	Data[0x124F] = (unsigned char) ((i >> 24) & 0xFF);
	i = SH2_Get_MACH(&M_SH2);
	Data[0x1250] = (unsigned char) (i & 0xFF);
	Data[0x1251] = (unsigned char) ((i >> 8) & 0xFF);
	Data[0x1252] = (unsigned char) ((i >> 16) & 0xFF);
	Data[0x1253] = (unsigned char) ((i >> 24) & 0xFF);
	i = SH2_Get_PR(&M_SH2);
	Data[0x1254] = (unsigned char) (i & 0xFF);
	Data[0x1255] = (unsigned char) ((i >> 8) & 0xFF);
	Data[0x1256] = (unsigned char) ((i >> 16) & 0xFF);
	Data[0x1257] = (unsigned char) ((i >> 24) & 0xFF);
	i = SH2_Get_PC(&M_SH2);
	Data[0x1258] = (unsigned char) (i & 0xFF);
	Data[0x1259] = (unsigned char) ((i >> 8) & 0xFF);
	Data[0x125A] = (unsigned char) ((i >> 16) & 0xFF);
	Data[0x125B] = (unsigned char) ((i >> 24) & 0xFF);
	i = M_SH2.SH2Status;
	Data[0x125C] = (unsigned char) (i & 0xFF);
	Data[0x125D] = (unsigned char) ((i >> 8) & 0xFF);
	Data[0x125E] = (unsigned char) ((i >> 16) & 0xFF);
	Data[0x125F] = (unsigned char) ((i >> 24) & 0xFF);

	for(i = 0; i < 0x1000; i++) Data[0x1400 + i] = S_SH2.Cache[i];
	for(i = 0; i < 0x200; i++) Data[0x2400 + i] = S_SH2.IO_Reg[i];

	src = (unsigned char *) &(S_SH2.R[0]);
	for(i = 0; i < 16 * 4; i++) Data[0x2600 + i] = *src++;

	i = SH2_Get_SR(&S_SH2);
	Data[0x2640] = (unsigned char) (i & 0xFF);
	Data[0x2641] = (unsigned char) ((i >> 8) & 0xFF);
	Data[0x2642] = (unsigned char) ((i >> 16) & 0xFF);
	Data[0x2643] = (unsigned char) ((i >> 24) & 0xFF);
	i = SH2_Get_GBR(&S_SH2);
	Data[0x2644] = (unsigned char) (i & 0xFF);
	Data[0x2645] = (unsigned char) ((i >> 8) & 0xFF);
	Data[0x2646] = (unsigned char) ((i >> 16) & 0xFF);
	Data[0x2647] = (unsigned char) ((i >> 24) & 0xFF);
	i = SH2_Get_VBR(&S_SH2);
	Data[0x2648] = (unsigned char) (i & 0xFF);
	Data[0x2649] = (unsigned char) ((i >> 8) & 0xFF);
	Data[0x264A] = (unsigned char) ((i >> 16) & 0xFF);
	Data[0x264B] = (unsigned char) ((i >> 24) & 0xFF);
	i = SH2_Get_MACL(&S_SH2);
	Data[0x264C] = (unsigned char) (i & 0xFF);
	Data[0x264D] = (unsigned char) ((i >> 8) & 0xFF);
	Data[0x264E] = (unsigned char) ((i >> 16) & 0xFF);
	Data[0x264F] = (unsigned char) ((i >> 24) & 0xFF);
	i = SH2_Get_MACH(&S_SH2);
	Data[0x2650] = (unsigned char) (i & 0xFF);
	Data[0x2651] = (unsigned char) ((i >> 8) & 0xFF);
	Data[0x2652] = (unsigned char) ((i >> 16) & 0xFF);
	Data[0x2653] = (unsigned char) ((i >> 24) & 0xFF);
	i = SH2_Get_PR(&S_SH2);
	Data[0x2654] = (unsigned char) (i & 0xFF);
	Data[0x2655] = (unsigned char) ((i >> 8) & 0xFF);
	Data[0x2656] = (unsigned char) ((i >> 16) & 0xFF);
	Data[0x2657] = (unsigned char) ((i >> 24) & 0xFF);
	i = SH2_Get_PC(&S_SH2);
	Data[0x2658] = (unsigned char) (i & 0xFF);
	Data[0x2659] = (unsigned char) ((i >> 8) & 0xFF);
	Data[0x265A] = (unsigned char) ((i >> 16) & 0xFF);
	Data[0x265B] = (unsigned char) ((i >> 24) & 0xFF);
	i = S_SH2.SH2Status;
	Data[0x265C] = (unsigned char) (i & 0xFF);
	Data[0x265D] = (unsigned char) ((i >> 8) & 0xFF);
	Data[0x265E] = (unsigned char) ((i >> 16) & 0xFF);
	Data[0x265F] = (unsigned char) ((i >> 24) & 0xFF);

	// Do it to allow VDP write on 32X side
	Data[0x2700] = _32X_ADEN & 1;
	Data[0x2705] = _32X_RV & 1;
	Data[0x2701] = _32X_FM & 0x80;
	Data[0x2702] = _32X_MINT;
	Data[0x2703] = _32X_SINT;
	Data[0x2704] = Bank_SH2;
	
	/*******
	FIFO stuff to add here...	
	 *******/
	for(i = 0; i < 0x10; i++) Data[0x2740 + i] = SH2_Read_Byte(&M_SH2, 0x4020 + i);
	for(i = 0; i < 4; i++) Data[0x2750 + i] = SH2_Read_Byte(&M_SH2, 0x4030 + i);
	/*******
	Extra PWM stuff to add here...	
	 *******/

	// Do it to allow VDP write on 32X side
	_32X_FM = 0x80;
	_32X_Set_FB();

	for(i = 0; i < 0x10; i++) Data[0x2760 + i] = SH2_Read_Byte(&M_SH2, 0x4100 + i);
	for(i = 0; i < 0x200; i += 2)
	{
		Data[0x2800 + i + 0] = (unsigned char) (SH2_Read_Word(&M_SH2, 0x4200 + i) & 0xFF);
		Data[0x2800 + i + 1] = (unsigned char) (SH2_Read_Word(&M_SH2, 0x4200 + i) >> 8);
	}
	for(i = 0; i < 0x40000; i++) Data[0x2A00 + i] = SH2_Read_Byte(&M_SH2, 0x6000000 + i);

	_32X_FM = Data[0x2701] & 0x80;
	_32X_Set_FB();

	if (SH2_Read_Word(&M_SH2, 0x410A) & 1)
	{
		for(i = 0; i < 0x20000; i++) Data[0x42A00 + i] = _32X_VDP_Ram[i];
		for(i = 0; i < 0x20000; i++) Data[0x62A00 + i] = _32X_VDP_Ram[i + 0x20000];
	}
	else
	{
		for(i = 0; i < 0x20000; i++) Data[0x42A00 + i] = _32X_VDP_Ram[i + 0x20000];
		for(i = 0; i < 0x20000; i++) Data[0x62A00 + i] = _32X_VDP_Ram[i];
	}
}


int Save_Config(char *File_Name)
{
	char Conf_File[1024];

	strcpy(Conf_File, File_Name);

    FILE* conf_fh = fopen(Conf_File,"w");
    fprintf(conf_fh,"[General]\n");
    fprintf(conf_fh,"Rom path=%s\n",Rom_Dir);
    fprintf(conf_fh,"Save Path=%s\n",State_Dir);
    fprintf(conf_fh,"SRAM path=%s\n",SRAM_Dir);
    fprintf(conf_fh,"BRAM path=%s\n",BRAM_Dir);
    fprintf(conf_fh,"Dump path=%s\n",Dump_Dir);
    fprintf(conf_fh,"Dump GYM path=%s\n",Dump_GYM_Dir);
    fprintf(conf_fh,"Screen Shot path=%s\n",ScrShot_Dir);
    fprintf(conf_fh,"Patch path=%s\n",Patch_Dir);
    fprintf(conf_fh,"IPS Patch path=%s\n",IPS_Dir);
    fprintf(conf_fh,"Genesis Bios=%s\n",Genesis_Bios);
    fprintf(conf_fh,"USA CD Bios=%s\n",US_CD_Bios);
    fprintf(conf_fh,"EUROPE CD Bios=%s\n",EU_CD_Bios);
    fprintf(conf_fh,"JAPAN CD Bios=%s\n",JA_CD_Bios);
    fprintf(conf_fh,"32X 68000 Bios=%s\n", _32X_Genesis_Bios);
    fprintf(conf_fh,"32X Master SH2 Bios=%s\n",_32X_Master_Bios);
    fprintf(conf_fh,"32X Slave SH2 Bios=%s\n",_32X_Slave_Bios);
    fprintf(conf_fh,"File type index=%d\n",File_Type_Index);
    fprintf(conf_fh,"State Number=%d\n",Current_State);
    fprintf(conf_fh,"Language=%d\n", Language);
    fprintf(conf_fh,"Window X=%d\n",WindowX);
    fprintf(conf_fh,"Window Y=%d\n",WindowY);
    fprintf(conf_fh,"Free Mode Color=%d\n",Effect_Color);

    fprintf(conf_fh,"\n[Graphics]\n");
    fprintf(conf_fh,"Full Screen=%d\n",Full_Screen & 1);
    fprintf(conf_fh,"Full Screen VSync=%d\n",FS_VSync & 1);
    fprintf(conf_fh,"Windows VSync=%d\n",W_VSync & 1);
    fprintf(conf_fh,"Render Windowed=%d\n",Render_W);
    fprintf(conf_fh,"Render Fullscreen=%d\n",Render_FS);
    fprintf(conf_fh,"Stretch=%d\n",Stretch & 1);
    fprintf(conf_fh,"Software Blit=%d\n",Blit_Soft & 1);
    fprintf(conf_fh,"Contrast=%d\n",Contrast_Level);
    fprintf(conf_fh,"Brightness=%d\n",Brightness_Level);
    fprintf(conf_fh,"Greyscale=%d\n",Greyscale & 1);
    fprintf(conf_fh,"Invert=%d\n",Invert_Color & 1);
    fprintf(conf_fh,"Sprite limit=%d\n",Sprite_Over & 1);
    fprintf(conf_fh,"Frame skip=%d\n",Frame_Skip);

    fprintf(conf_fh,"\n[Sound]\n");
    fprintf(conf_fh,"State=%d\n",Sound_Enable & 1);
    fprintf(conf_fh,"Rate=%d\n",Sound_Rate);
    fprintf(conf_fh,"Stereo=%d\n",Sound_Stereo);
    fprintf(conf_fh,"Z80 State=%d\n", Z80_State & 1);
    fprintf(conf_fh,"YM2612 State=%d\n",YM2612_Enable & 1);
    fprintf(conf_fh,"PSG State=%d\n",PSG_Enable & 1);
    fprintf(conf_fh,"DAC State=%d\n",DAC_Enable & 1);
    fprintf(conf_fh,"PCM State=%d\n",PCM_Enable & 1);
    fprintf(conf_fh,"PWM State=%d\n",PWM_Enable & 1);
    fprintf(conf_fh,"CDDA State=%d\n",CDDA_Enable & 1);
    fprintf(conf_fh,"YM2612 Improvement=%d\n",YM2612_Improv & 1);
    fprintf(conf_fh,"DAC Improvement=%d\n", DAC_Improv & 1);
    fprintf(conf_fh,"PSG Improvement=%d\n",PSG_Improv & 1);
    
    fprintf(conf_fh,"\n[CPU]\n");
    fprintf(conf_fh,"Country=%d\n",Country);
    fprintf(conf_fh,"Prefered Country 1=%d\n", Country_Order[0]);
    fprintf(conf_fh,"Prefered Country 2=%d\n", Country_Order[1]);
    fprintf(conf_fh,"Prefered Country 3=%d\n", Country_Order[2]);
    fprintf(conf_fh,"Perfect synchro between main and sub CPU (Sega CD)=%d\n",SegaCD_Accurate);
    fprintf(conf_fh,"Main SH2 Speed=%d\n",MSH2_Speed);
    fprintf(conf_fh,"Slave SH2 Speed=%d\n",SSH2_Speed);

    fprintf(conf_fh,"\n[Options]\n");
    fprintf(conf_fh,"Fast Blur=%d\n",Fast_Blur & 1);
    fprintf(conf_fh,"FPS=%d\n",Show_FPS & 1);
    fprintf(conf_fh,"FPS Style=%d\n",FPS_Style);
    fprintf(conf_fh,"Message=%d\n",Show_Message & 1);
    fprintf(conf_fh,"Message Style=%d\n",Message_Style);
    fprintf(conf_fh,"LED=%d\n",Show_LED & 1);
    fprintf(conf_fh,"Auto Fix Checksum=%d\n",Auto_Fix_CS & 1);
    fprintf(conf_fh,"Auto Pause=%d\n",Auto_Pause & 1);
    fprintf(conf_fh,"CD Drive=%d\n",CUR_DEV);

	if (BRAM_Ex_State & 0x100)
        fprintf(conf_fh,"Ram Cart Size=%d\n",BRAM_Ex_Size);
	else
        fprintf(conf_fh,"Ram Cart Size=-1\n");

    fprintf(conf_fh,"Disable Blue Screen=%d\n",DisableBlueScreen);
    fprintf(conf_fh,"Enable Frame Counter=%d\n",FrameCounterEnabled);

    if (FrameCounterPosition == 5650) // tr
        fprintf(conf_fh,"Frame Counter Position=tr\n");
    else if (FrameCounterPosition == 67216) // bl
        fprintf(conf_fh,"Frame Counter Position=bl\n");
    else if (FrameCounterPosition == 67470) // br
        fprintf(conf_fh,"Frame Counter Position=br\n");
    else if (FrameCounterPosition == 5392) // tl
        fprintf(conf_fh,"Frame Counter Position=tl\n");
    else
        fprintf(conf_fh,"Frame Counter Position=%d\n",FrameCounterPosition);

    fprintf(conf_fh,"Show Input=%d\n",ShowInputEnabled);
    fprintf(conf_fh,"Auto-backup=%d\n",AutoBackupEnabled);

	/*WritePrivateProfileString("Options", "GCOffline path", CGOffline_Path, Conf_File);
	WritePrivateProfileString("Options", "Gens manual path", Manual_Path, Conf_File);*/

    fprintf(conf_fh,"\n[Input]\n");
    fprintf(conf_fh,"P1.Type=%d\n",Controller_1_Type & 0x13);
    WriteSDLKey(conf_fh,"P1.Up",Keys_Def[0].Up);
    WriteSDLKey(conf_fh,"P1.Down",Keys_Def[0].Down);
    WriteSDLKey(conf_fh,"P1.Left",Keys_Def[0].Left);
    WriteSDLKey(conf_fh,"P1.Right",Keys_Def[0].Right);
    WriteSDLKey(conf_fh,"P1.Start",Keys_Def[0].Start);
    WriteSDLKey(conf_fh,"P1.A",Keys_Def[0].A);
    WriteSDLKey(conf_fh,"P1.B",Keys_Def[0].B);
    WriteSDLKey(conf_fh,"P1.C",Keys_Def[0].C);
    WriteSDLKey(conf_fh,"P1.Mode",Keys_Def[0].Mode);
    WriteSDLKey(conf_fh,"P1.X",Keys_Def[0].X);
    WriteSDLKey(conf_fh,"P1.Y",Keys_Def[0].Y);
    WriteSDLKey(conf_fh,"P1.Z",Keys_Def[0].Z);

    fprintf(conf_fh,"P1B.Type=%d\n",Controller_1B_Type & 0x03);
    WriteSDLKey(conf_fh,"P1B.Up",Keys_Def[2].Up);
    WriteSDLKey(conf_fh,"P1B.Down",Keys_Def[2].Down);
    WriteSDLKey(conf_fh,"P1B.Left",Keys_Def[2].Left);
    WriteSDLKey(conf_fh,"P1B.Right",Keys_Def[2].Right);
    WriteSDLKey(conf_fh,"P1B.Start",Keys_Def[2].Start);
    WriteSDLKey(conf_fh,"P1B.A",Keys_Def[2].A);
    WriteSDLKey(conf_fh,"P1B.B",Keys_Def[2].B);
    WriteSDLKey(conf_fh,"P1B.C",Keys_Def[2].C);
    WriteSDLKey(conf_fh,"P1B.Mode",Keys_Def[2].Mode);
    WriteSDLKey(conf_fh,"P1B.X",Keys_Def[2].X);
    WriteSDLKey(conf_fh,"P1B.Y",Keys_Def[2].Y);
    WriteSDLKey(conf_fh,"P1B.Z",Keys_Def[2].Z);

    fprintf(conf_fh,"P1C.Type=%d\n",Controller_1C_Type & 0x13);
    WriteSDLKey(conf_fh,"P1C.Up",Keys_Def[3].Up);
    WriteSDLKey(conf_fh,"P1C.Down",Keys_Def[3].Down);
    WriteSDLKey(conf_fh,"P1C.Left",Keys_Def[3].Left);
    WriteSDLKey(conf_fh,"P1C.Right",Keys_Def[3].Right);
    WriteSDLKey(conf_fh,"P1C.Start",Keys_Def[3].Start);
    WriteSDLKey(conf_fh,"P1C.A",Keys_Def[3].A);
    WriteSDLKey(conf_fh,"P1C.B",Keys_Def[3].B);
    WriteSDLKey(conf_fh,"P1C.C",Keys_Def[3].C);
    WriteSDLKey(conf_fh,"P1C.Mode",Keys_Def[3].Mode);
    WriteSDLKey(conf_fh,"P1C.X",Keys_Def[3].X);
    WriteSDLKey(conf_fh,"P1C.Y",Keys_Def[3].Y);
    WriteSDLKey(conf_fh,"P1C.Z",Keys_Def[3].Z);

    fprintf(conf_fh,"P1D.Type=%d\n",Controller_1D_Type & 0x03);
    WriteSDLKey(conf_fh,"P1D.Up",Keys_Def[4].Up);
    WriteSDLKey(conf_fh,"P1D.Down",Keys_Def[4].Down);
    WriteSDLKey(conf_fh,"P1D.Left",Keys_Def[4].Left);
    WriteSDLKey(conf_fh,"P1D.Right",Keys_Def[4].Right);
    WriteSDLKey(conf_fh,"P1D.Start",Keys_Def[4].Start);
    WriteSDLKey(conf_fh,"P1D.A",Keys_Def[4].A);
    WriteSDLKey(conf_fh,"P1D.B",Keys_Def[4].B);
    WriteSDLKey(conf_fh,"P1D.C",Keys_Def[4].C);
    WriteSDLKey(conf_fh,"P1D.Mode",Keys_Def[4].Mode);
    WriteSDLKey(conf_fh,"P1D.X",Keys_Def[4].X);
    WriteSDLKey(conf_fh,"P1D.Y",Keys_Def[4].Y);
    WriteSDLKey(conf_fh,"P1D.Z",Keys_Def[4].Z);

    fprintf(conf_fh,"P2.Type=%d\n",Controller_2_Type & 0x13);
    WriteSDLKey(conf_fh,"P2.Up",Keys_Def[1].Up);
    WriteSDLKey(conf_fh,"P2.Down",Keys_Def[1].Down);
    WriteSDLKey(conf_fh,"P2.Left",Keys_Def[1].Left);
    WriteSDLKey(conf_fh,"P2.Right",Keys_Def[1].Right);
    WriteSDLKey(conf_fh,"P2.Start",Keys_Def[1].Start);
    WriteSDLKey(conf_fh,"P2.A",Keys_Def[1].A);
    WriteSDLKey(conf_fh,"P2.B",Keys_Def[1].B);
    WriteSDLKey(conf_fh,"P2.C",Keys_Def[1].C);
    WriteSDLKey(conf_fh,"P2.Mode",Keys_Def[1].Mode);
    WriteSDLKey(conf_fh,"P2.X",Keys_Def[1].X);
    WriteSDLKey(conf_fh,"P2.Y",Keys_Def[1].Y);
    WriteSDLKey(conf_fh,"P2.Z",Keys_Def[1].Z);

    fprintf(conf_fh,"P2B.Type=%d\n",Controller_2B_Type & 0x13);
    WriteSDLKey(conf_fh,"P2B.Up",Keys_Def[5].Up);
    WriteSDLKey(conf_fh,"P2B.Down",Keys_Def[5].Down);
    WriteSDLKey(conf_fh,"P2B.Left",Keys_Def[5].Left);
    WriteSDLKey(conf_fh,"P2B.Right",Keys_Def[5].Right);
    WriteSDLKey(conf_fh,"P2B.Start",Keys_Def[5].Start);
    WriteSDLKey(conf_fh,"P2B.A",Keys_Def[5].A);
    WriteSDLKey(conf_fh,"P2B.B",Keys_Def[5].B);
    WriteSDLKey(conf_fh,"P2B.C",Keys_Def[5].C);
    WriteSDLKey(conf_fh,"P2B.Mode",Keys_Def[5].Mode);
    WriteSDLKey(conf_fh,"P2B.X",Keys_Def[5].X);
    WriteSDLKey(conf_fh,"P2B.Y",Keys_Def[5].Y);
    WriteSDLKey(conf_fh,"P2B.Z",Keys_Def[5].Z);

    fprintf(conf_fh,"P2C.Type=%d\n",Controller_2C_Type & 0x03);
    WriteSDLKey(conf_fh,"P2C.Up",Keys_Def[1].Up);
    WriteSDLKey(conf_fh,"P2C.Down",Keys_Def[6].Down);
    WriteSDLKey(conf_fh,"P2C.Left",Keys_Def[6].Left);
    WriteSDLKey(conf_fh,"P2C.Right",Keys_Def[6].Right);
    WriteSDLKey(conf_fh,"P2C.Start",Keys_Def[6].Start);
    WriteSDLKey(conf_fh,"P2C.A",Keys_Def[6].A);
    WriteSDLKey(conf_fh,"P2C.B",Keys_Def[6].B);
    WriteSDLKey(conf_fh,"P2C.C",Keys_Def[6].C);
    WriteSDLKey(conf_fh,"P2C.Mode",Keys_Def[6].Mode);
    WriteSDLKey(conf_fh,"P2C.X",Keys_Def[6].X);
    WriteSDLKey(conf_fh,"P2C.Y",Keys_Def[6].Y);
    WriteSDLKey(conf_fh,"P2C.Z",Keys_Def[6].Z);

    fprintf(conf_fh,"P2D.Type=%d\n",Controller_2D_Type & 0x03);
    WriteSDLKey(conf_fh,"P2D.Up",Keys_Def[7].Up);
    WriteSDLKey(conf_fh,"P2D.Down",Keys_Def[7].Down);
    WriteSDLKey(conf_fh,"P2D.Left",Keys_Def[7].Left);
    WriteSDLKey(conf_fh,"P2D.Right",Keys_Def[7].Right);
    WriteSDLKey(conf_fh,"P2D.Start",Keys_Def[7].Start);
    WriteSDLKey(conf_fh,"P2D.A",Keys_Def[7].A);
    WriteSDLKey(conf_fh,"P2D.B",Keys_Def[7].B);
    WriteSDLKey(conf_fh,"P2D.C",Keys_Def[7].C);
    WriteSDLKey(conf_fh,"P2D.Mode",Keys_Def[7].Mode);
    WriteSDLKey(conf_fh,"P2D.X",Keys_Def[7].X);
    WriteSDLKey(conf_fh,"P2D.Y",Keys_Def[7].Y);
    WriteSDLKey(conf_fh,"P2D.Z",Keys_Def[7].Z);

	/*
	sprintf(Str_Tmp, "%d", Controller_2B_Type & 0x03);
	WritePrivateProfileString("Input", "P2B.Type", Str_Tmp, Conf_File);
	sprintf(Str_Tmp, "%d", Keys_Def[5].Up);

	sprintf(Str_Tmp, "%d", Controller_2C_Type & 0x03);
	WritePrivateProfileString("Input", "P2C.Type", Str_Tmp, Conf_File);
	sprintf(Str_Tmp, "%d", Keys_Def[6].Up);

	sprintf(Str_Tmp, "%d", Controller_2D_Type & 0x03);
	WritePrivateProfileString("Input", "P2D.Type", Str_Tmp, Conf_File);
	sprintf(Str_Tmp, "%d", Keys_Def[7].Up);*/

    fprintf(conf_fh,"\n[Interface]\n");
    WriteSDLkeysym(conf_fh,"Pause Emulator",Pause_Emulator_key);
	WriteSDLkeysym(conf_fh,"Quit",Quit_key);
	WriteSDLkeysym(conf_fh,"Take Screenshot",Take_Screenshot_key);
	WriteSDLkeysym(conf_fh,"Fast Forward",Fast_Forward_key);
	WriteSDLkeysym(conf_fh,"Reset",Reset_key);
	WriteSDLkeysym(conf_fh,"Lower Speed",Lower_Speed_key);
	WriteSDLkeysym(conf_fh,"Raise Speed",Raise_Speed_key);
	WriteSDLkeysym(conf_fh,"Toggle Fullscreen",Toggle_Fullscreen_key);
	WriteSDLkeysym(conf_fh,"Change Stretch",Change_Stretch_key);
	WriteSDLkeysym(conf_fh,"Auto Frameskip",Auto_Frameskip_key);
	WriteSDLkeysym(conf_fh,"Change Vsync",Change_Vsync_key);
	WriteSDLkeysym(conf_fh,"Lower Frameskip",Lower_Frameskip_key);
	WriteSDLkeysym(conf_fh,"Raise Frameskip",Raise_Frameskip_key);
	WriteSDLkeysym(conf_fh,"Save Current State",Save_Current_State_key);
	WriteSDLkeysym(conf_fh,"Alternate Save Current State",Alt_Save_Current_State_key);
	WriteSDLkeysym(conf_fh,"Select Previous State",Select_Previous_State_key);
	WriteSDLkeysym(conf_fh,"Select Next State",Select_Next_State_key);
	WriteSDLkeysym(conf_fh,"Load Current State",Load_Current_State_key);
	WriteSDLkeysym(conf_fh,"Alternate Load Current State",Alt_Load_Current_State_key);
	WriteSDLkeysym(conf_fh,"Change Blit Style",Change_Blit_Style_key);
	WriteSDLkeysym(conf_fh,"Toggle Fast Blur",Toggle_Fast_Blur_key);
	WriteSDLkeysym(conf_fh,"Toggle Improved DAC",Toggle_Improved_DAC_key);
	WriteSDLkeysym(conf_fh,"Toggle FPS Display",Toggle_FPS_Display_key);
	WriteSDLkeysym(conf_fh,"Toggle Improved PSG",Toggle_Improved_PSG_key);
	WriteSDLkeysym(conf_fh,"Previous Rendering Mode",Previous_Rendering_Mode_key);
	WriteSDLkeysym(conf_fh,"Toggle Improved YM2612",Toggle_Improved_YM2612_key);
	WriteSDLkeysym(conf_fh,"Next Rendering Mode",Next_Rendering_Mode_key);
	WriteSDLkeysym(conf_fh,"Select State 0",Select_State_0_key);
	WriteSDLkeysym(conf_fh,"Select State 1",Select_State_1_key);
	WriteSDLkeysym(conf_fh,"Select State 2",Select_State_2_key);
	WriteSDLkeysym(conf_fh,"Select State 3",Select_State_3_key);
	WriteSDLkeysym(conf_fh,"Select State 4",Select_State_4_key);
	WriteSDLkeysym(conf_fh,"Select State 5",Select_State_5_key);
	WriteSDLkeysym(conf_fh,"Select State 6",Select_State_6_key);
	WriteSDLkeysym(conf_fh,"Select State 7",Select_State_7_key);
	WriteSDLkeysym(conf_fh,"Select State 8",Select_State_8_key);
	WriteSDLkeysym(conf_fh,"Select State 9",Select_State_9_key);
	WriteSDLkeysym(conf_fh,"Load State 0",Load_State_0_key);
	WriteSDLkeysym(conf_fh,"Load State 1",Load_State_1_key);
	WriteSDLkeysym(conf_fh,"Load State 2",Load_State_2_key);
	WriteSDLkeysym(conf_fh,"Load State 3",Load_State_3_key);
	WriteSDLkeysym(conf_fh,"Load State 4",Load_State_4_key);
	WriteSDLkeysym(conf_fh,"Load State 5",Load_State_5_key);
	WriteSDLkeysym(conf_fh,"Load State 6",Load_State_6_key);
	WriteSDLkeysym(conf_fh,"Load State 7",Load_State_7_key);
	WriteSDLkeysym(conf_fh,"Load State 8",Load_State_8_key);
	WriteSDLkeysym(conf_fh,"Load State 9",Load_State_9_key);
	WriteSDLkeysym(conf_fh,"Save State 0",Save_State_0_key);
	WriteSDLkeysym(conf_fh,"Save State 1",Save_State_1_key);
	WriteSDLkeysym(conf_fh,"Save State 2",Save_State_2_key);
	WriteSDLkeysym(conf_fh,"Save State 3",Save_State_3_key);
	WriteSDLkeysym(conf_fh,"Save State 4",Save_State_4_key);
	WriteSDLkeysym(conf_fh,"Save State 5",Save_State_5_key);
	WriteSDLkeysym(conf_fh,"Save State 6",Save_State_6_key);
	WriteSDLkeysym(conf_fh,"Save State 7",Save_State_7_key);
	WriteSDLkeysym(conf_fh,"Save State 8",Save_State_8_key);
	WriteSDLkeysym(conf_fh,"Save State 9",Save_State_9_key);
	WriteSDLkeysym(conf_fh,"Close ROM",Close_ROM_key);
	WriteSDLkeysym(conf_fh,"Open Game Genie",Open_Game_Genie_key);
	WriteSDLkeysym(conf_fh,"Change SegaCD Syncro",Change_SegaCD_Syncro_key);
	WriteSDLkeysym(conf_fh,"Start Playing GYM",Start_Playing_GYM_key);
	WriteSDLkeysym(conf_fh,"Change CD",Change_CD_key);
	WriteSDLkeysym(conf_fh,"Toggle WAV Dumping",Toggle_WAV_Dumping_key);
	WriteSDLkeysym(conf_fh,"RAM Search Menu",RAM_Search_Menu_key);
	WriteSDLkeysym(conf_fh,"Start Recording",Start_Recording_key);
	WriteSDLkeysym(conf_fh,"Toggle Readonly",Toggle_Readonly_key);
	WriteSDLkeysym(conf_fh,"Play Movie",Play_Movie_key);
	WriteSDLkeysym(conf_fh,"Toggle Frame Counter",Toggle_Frame_Counter_key);
	WriteSDLkeysym(conf_fh,"Advance Frame",Advance_Frame_key);
	WriteSDLkeysym(conf_fh,"Toggle Input Display",Toggle_Input_Display_key);
	WriteSDLkeysym(conf_fh,"Toggle Sound",Toggle_Sound_key);
	WriteSDLkeysym(conf_fh,"Stop Movie",Stop_Movie_key);
	WriteSDLkeysym(conf_fh,"Record New Movie",Record_New_Movie_key);
	WriteSDLkeysym(conf_fh,"Load State File",Load_State_File_key);
	WriteSDLkeysym(conf_fh,"Save State File",Save_State_File_key);
	WriteSDLkeysym(conf_fh,"Save Config File",Save_Config_File_key); // */

    fclose(conf_fh);
	return 1;
}


int Save_As_Config(HWND hWnd)
{
#ifdef __PORT__
#ifdef WITH_GTK
	char Name[2048];
	GtkWidget* filesel;
	gint res;

	filesel = create_file_selector_private(NULL/*Filter*/,0);	
	res = gtk_dialog_run(GTK_DIALOG(filesel));
	
	switch(res)
	{
		case GTK_RESPONSE_OK:
			strcpy(Name,gtk_file_selection_get_filename(GTK_FILE_SELECTION(filesel)));
			break;
		case GTK_RESPONSE_CANCEL:
			break;
	}

	gtk_widget_destroy(filesel);
	
#endif	
#else
	char Name[2048];
	OPENFILENAME ofn;

	SetCurrentDirectory(Gens_Path);

	strcpy(&Name[0], "Gens.cfg");
	memset(&ofn, 0, sizeof(OPENFILENAME));

	ofn.lStructSize = sizeof(OPENFILENAME);
	ofn.hwndOwner = hWnd;
	ofn.hInstance = ghInstance;
	ofn.lpstrFile = Name;
	ofn.nMaxFile = 2047;
	ofn.lpstrFilter = "Config Files\0*.cfg\0All Files\0*.*\0\0";
	ofn.nFilterIndex = 1;
	ofn.lpstrInitialDir = Gens_Path;
	ofn.lpstrDefExt = "cfg";
	ofn.Flags = 0;
	
#endif
#ifdef __PORT__
#ifdef WITH_GTK
	if (strlen(Name))
	{
		Save_Config(Name);
		strcpy(Str_Tmp, "config saved in ");
		strcat(Str_Tmp, Name);
		Put_Info(Str_Tmp, 2000);
		return 1;	
	}
#endif
#else
	if(GetSaveFileName(&ofn))
	{
		Save_Config(Name);
		strcpy(Str_Tmp, "config saved in ");
		strcat(Str_Tmp, Name);
		Put_Info(Str_Tmp, 2000);
		return 1;	
	}
#endif
	return 0;
}


int Load_Config(char *File_Name, void *Game_Active)
{
	int new_val;
	char Conf_File[1024];

	SetCurrentDirectory(Gens_Path);
	strcpy(Conf_File, File_Name);

	CRam_Flag = 1;

	/*GetPrivateProfileString("General", "Rom path", default_path, &Rom_Dir[0], 1024, Conf_File);
	GetPrivateProfileString("General", "Save path", Rom_Dir, &State_Dir[0], 1024, Conf_File);
	GetPrivateProfileString("General", "SRAM path", Rom_Dir, &SRAM_Dir[0], 1024, Conf_File);
	GetPrivateProfileString("General", "BRAM path", Rom_Dir, &BRAM_Dir[0], 1024, Conf_File);
	GetPrivateProfileString("General", "Dump path", Rom_Dir, &Dump_Dir[0], 1024, Conf_File);
	GetPrivateProfileString("General", "Dump GYM path", Rom_Dir, &Dump_GYM_Dir[0], 1024, Conf_File);
	GetPrivateProfileString("General", "Screen Shot path", Rom_Dir, &ScrShot_Dir[0], 1024, Conf_File);
	GetPrivateProfileString("General", "Patch path", Rom_Dir, &Patch_Dir[0], 1024, Conf_File);
	GetPrivateProfileString("General", "IPS Patch path", Rom_Dir, &IPS_Dir[0], 1024, Conf_File);*/

	GetPrivateProfileString("General", "Genesis Bios", Rom_Dir, &Genesis_Bios[0], 1024, Conf_File);

	GetPrivateProfileString("General", "USA CD Bios", Rom_Dir, &US_CD_Bios[0], 1024, Conf_File);
	GetPrivateProfileString("General", "EUROPE CD Bios", Rom_Dir, &EU_CD_Bios[0], 1024, Conf_File);
	GetPrivateProfileString("General", "JAPAN CD Bios", Rom_Dir, &JA_CD_Bios[0], 1024, Conf_File);

	GetPrivateProfileString("General", "32X 68000 Bios", Rom_Dir, &_32X_Genesis_Bios[0], 1024, Conf_File);
	GetPrivateProfileString("General", "32X Master SH2 Bios", Rom_Dir, &_32X_Master_Bios[0], 1024, Conf_File);
	GetPrivateProfileString("General", "32X Slave SH2 Bios", Rom_Dir, &_32X_Slave_Bios[0], 1024, Conf_File);

	/*GetPrivateProfileString("General", "Rom 1", "", &Recent_Rom[0][0], 1024, Conf_File);
	GetPrivateProfileString("General", "Rom 2", "", &Recent_Rom[1][0], 1024, Conf_File);
	GetPrivateProfileString("General", "Rom 3", "", &Recent_Rom[2][0], 1024, Conf_File);
	GetPrivateProfileString("General", "Rom 4", "", &Recent_Rom[3][0], 1024, Conf_File);
	GetPrivateProfileString("General", "Rom 5", "", &Recent_Rom[4][0], 1024, Conf_File);
	GetPrivateProfileString("General", "Rom 6", "", &Recent_Rom[5][0], 1024, Conf_File);
	GetPrivateProfileString("General", "Rom 7", "", &Recent_Rom[6][0], 1024, Conf_File);
	GetPrivateProfileString("General", "Rom 8", "", &Recent_Rom[7][0], 1024, Conf_File);
	GetPrivateProfileString("General", "Rom 9", "", &Recent_Rom[8][0], 1024, Conf_File);*/

	File_Type_Index = GetPrivateProfileInt("General", "File type index", 1, Conf_File);

	Current_State = GetPrivateProfileInt("General", "State Number", 0, Conf_File);
	Language = GetPrivateProfileInt("General", "Language", 0, Conf_File);
    //Modif IM - works for non-win32 now
	WindowX = GetPrivateProfileInt("General", "Window X", 0, Conf_File);
	WindowY = GetPrivateProfileInt("General", "Window Y", 0, Conf_File);
	Intro_Style = GetPrivateProfileInt("General", "Intro Style", 0, Conf_File);
	Effect_Color = GetPrivateProfileInt("General", "Free Mode Color", 7, Conf_File);
	Sleep_Time = GetPrivateProfileInt("General", "Allow Idle", 0, Conf_File) & 1;
	Gens_Priority = GetPrivateProfileInt("General", "Priority", 1, Conf_File);

	if (GetPrivateProfileInt("Graphics", "Force 555", 0, Conf_File)) Mode_555 = 3;
	else if (GetPrivateProfileInt("Graphics", "Force 565", 0, Conf_File)) Mode_555 = 2;
	else Mode_555 = 0;

	RMax_Level = GetPrivateProfileInt("Graphics", "Red Max", 255, Conf_File);
	GMax_Level = GetPrivateProfileInt("Graphics", "Green Max", 255, Conf_File);
	BMax_Level = GetPrivateProfileInt("Graphics", "Blue Max", 255, Conf_File);
	Contrast_Level = GetPrivateProfileInt("Graphics", "Contrast", 100, Conf_File);
	Brightness_Level = GetPrivateProfileInt("Graphics", "Brightness", 100, Conf_File);
	Greyscale = GetPrivateProfileInt("Graphics", "Greyscale", 0, Conf_File);
	Invert_Color = GetPrivateProfileInt("Graphics", "Invert", 0, Conf_File);

	Recalculate_Palettes();

	FS_VSync = GetPrivateProfileInt("Graphics", "Full Screen VSync", 0, Conf_File);
	W_VSync = GetPrivateProfileInt("Graphics", "Windows VSync", 0, Conf_File);
	Full_Screen = GetPrivateProfileInt("Graphics", "Full Screen", 0, Conf_File);
	Render_W = GetPrivateProfileInt("Graphics", "Render Windowed", 0, Conf_File);
	Render_FS = GetPrivateProfileInt("Graphics", "Render Fullscreen", 1, Conf_File);

	Set_Render(HWnd, Full_Screen, -1, 1);

	Stretch = GetPrivateProfileInt("Graphics", "Stretch", 0, Conf_File);
	Blit_Soft = GetPrivateProfileInt("Graphics", "Software Blit", 0, Conf_File);
	Sprite_Over = GetPrivateProfileInt("Graphics", "Sprite limit", 1, Conf_File);
	Frame_Skip = GetPrivateProfileInt("Graphics", "Frame skip", -1, Conf_File);

	Sound_Rate = GetPrivateProfileInt("Sound", "Rate", 22050, Conf_File);
	Sound_Stereo = GetPrivateProfileInt("Sound", "Stereo", 1, Conf_File);

	if (GetPrivateProfileInt("Sound", "Z80 State", 1, Conf_File)) Z80_State |= 1;
	else Z80_State &= ~1;

	new_val = GetPrivateProfileInt("Sound", "State", 1, Conf_File);
	if (new_val != Sound_Enable)
	{
//#ifndef __PORT__
		if (Change_Sound(HWnd))
//#endif
		{
			YM2612_Enable = GetPrivateProfileInt("Sound", "YM2612 State", 1, Conf_File);
			PSG_Enable = GetPrivateProfileInt("Sound", "PSG State", 1, Conf_File);
			DAC_Enable = GetPrivateProfileInt("Sound", "DAC State", 1, Conf_File);
			PCM_Enable = GetPrivateProfileInt("Sound", "PCM State", 1, Conf_File);
			PWM_Enable = GetPrivateProfileInt("Sound", "PWM State", 1, Conf_File);
			CDDA_Enable = GetPrivateProfileInt("Sound", "CDDA State", 1, Conf_File);
		}
	}
	else
	{
		YM2612_Enable = GetPrivateProfileInt("Sound", "YM2612 State", 1, Conf_File);
		PSG_Enable = GetPrivateProfileInt("Sound", "PSG State", 1, Conf_File);
		DAC_Enable = GetPrivateProfileInt("Sound", "DAC State", 1, Conf_File);
		PCM_Enable = GetPrivateProfileInt("Sound", "PCM State", 1, Conf_File);
		PWM_Enable = GetPrivateProfileInt("Sound", "PWM State", 1, Conf_File);
		CDDA_Enable = GetPrivateProfileInt("Sound", "CDDA State", 1, Conf_File);
	}

	YM2612_Improv = GetPrivateProfileInt("Sound", "YM2612 Improvement", 0, Conf_File);
	DAC_Improv = GetPrivateProfileInt("Sound", "DAC Improvement", 0, Conf_File);
	PSG_Improv = GetPrivateProfileInt("Sound", "PSG Improvement", 0, Conf_File);

	Country = GetPrivateProfileInt("CPU", "Country", -1, Conf_File);
	Country_Order[0] = GetPrivateProfileInt("CPU", "Prefered Country 1", 0, Conf_File);
	Country_Order[1] = GetPrivateProfileInt("CPU", "Prefered Country 2", 1, Conf_File);
	Country_Order[2] = GetPrivateProfileInt("CPU", "Prefered Country 3", 2, Conf_File);

	SegaCD_Accurate = GetPrivateProfileInt("CPU", "Perfect synchro between main and sub CPU (Sega CD)", 0, Conf_File);

	MSH2_Speed = GetPrivateProfileInt("CPU", "Main SH2 Speed", 100, Conf_File);
	SSH2_Speed = GetPrivateProfileInt("CPU", "Slave SH2 Speed", 100, Conf_File);

	if (MSH2_Speed < 0) MSH2_Speed = 0;
	if (SSH2_Speed < 0) SSH2_Speed = 0;

	Check_Country_Order();

	Fast_Blur = GetPrivateProfileInt("Options", "Fast Blur", 0, Conf_File);
	Show_FPS = GetPrivateProfileInt("Options", "FPS", 0, Conf_File);
	FPS_Style = GetPrivateProfileInt("Options", "FPS Style", 0, Conf_File);
	Show_Message = GetPrivateProfileInt("Options", "Message", 1, Conf_File);
	Message_Style = GetPrivateProfileInt("Options", "Message Style", 0, Conf_File);
	Show_LED = GetPrivateProfileInt("Options", "LED", 1, Conf_File);
	Auto_Fix_CS = GetPrivateProfileInt("Options", "Auto Fix Checksum", 0, Conf_File);
	Auto_Pause = GetPrivateProfileInt("Options", "Auto Pause", 0, Conf_File);
	CUR_DEV = GetPrivateProfileInt("Options", "CD Drive", 0, Conf_File);
	BRAM_Ex_Size = GetPrivateProfileInt("Options", "Ram Cart Size", 3, Conf_File);
    AutoBackupEnabled = GetPrivateProfileInt("Options", "Auto-backup",1,Conf_File);
    DisableBlueScreen = GetPrivateProfileInt("Options", "Disable Blue Screen",0,Conf_File);
    FrameCounterEnabled = GetPrivateProfileInt("Options", "Enable Frame Counter",1,Conf_File);
    ShowInputEnabled = GetPrivateProfileInt("Options", "Show Input",1,Conf_File);
    GetPrivateProfileString("Options", "Frame Counter Position","ul",Str_Tmp,2,Conf_File);
    if (!strncmp(Str_Tmp,"tr",2))
        FrameCounterPosition = 5650;
    else if (!strncmp(Str_Tmp,"bl",2))
        FrameCounterPosition = 67216;
    else if (!strncmp(Str_Tmp,"br",2))
        FrameCounterPosition = 67470;
    else if (!strncmp(Str_Tmp,"tl",2))
        FrameCounterPosition = 5392;
    else
        FrameCounterPosition = atoi(Str_Tmp);

	if (BRAM_Ex_Size == -1)
	{
		BRAM_Ex_State &= 1;
		BRAM_Ex_Size = 0;
	}
	else BRAM_Ex_State |= 0x100;

	GetPrivateProfileString("Options", "GCOffline path", "GCOffline.chm", CGOffline_Path, 1024, Conf_File);
	GetPrivateProfileString("Options", "Gens manual path", "manual.exe", Manual_Path, 1024, Conf_File);

    //Modif IM - make game keys configurable
	Controller_1_Type = GetPrivateProfileInt("Input", "P1.Type", 1, Conf_File);
    SetInputKey("P1.Up","[5]",Keys_Def[0].Up);
    SetInputKey("P1.Down","[2]",Keys_Def[0].Down);
    SetInputKey("P1.Left","[1]",Keys_Def[0].Left);
    SetInputKey("P1.Right","[3]",Keys_Def[0].Right);
    SetInputKey("P1.Start","space",Keys_Def[0].Start);
    SetInputKey("P1.A","t",Keys_Def[0].A);
    SetInputKey("P1.B","y",Keys_Def[0].B);
    SetInputKey("P1.C","k",Keys_Def[0].C);
    SetInputKey("P1.Mode","none",Keys_Def[0].Mode);
    SetInputKey("P1.X","none",Keys_Def[0].X);
    SetInputKey("P1.Y","none",Keys_Def[0].Y);
    SetInputKey("P1.Z","none",Keys_Def[0].Z); 

	Controller_1B_Type = GetPrivateProfileInt("Input", "P1B.Type", 0, Conf_File);
    SetInputKey("P1B.Up","none",Keys_Def[2].Up);
    SetInputKey("P1B.Down","none",Keys_Def[2].Down);
    SetInputKey("P1B.Left","none",Keys_Def[2].Left);
    SetInputKey("P1B.Right","none",Keys_Def[2].Right);
    SetInputKey("P1B.Start","none",Keys_Def[2].Start);
    SetInputKey("P1B.A","none",Keys_Def[2].A);
    SetInputKey("P1B.B","none",Keys_Def[2].B);
    SetInputKey("P1B.C","none",Keys_Def[2].C);
    SetInputKey("P1B.Mode","none",Keys_Def[2].Mode);
    SetInputKey("P1B.X","none",Keys_Def[2].X);
    SetInputKey("P1B.Y","none",Keys_Def[2].Y);
    SetInputKey("P1B.Z","none",Keys_Def[2].Z); 

	Controller_1C_Type = GetPrivateProfileInt("Input", "P1C.Type", 0, Conf_File);
    SetInputKey("P1C.Up","none",Keys_Def[3].Up);
    SetInputKey("P1C.Down","none",Keys_Def[3].Down);
    SetInputKey("P1C.Left","none",Keys_Def[3].Left);
    SetInputKey("P1C.Right","none",Keys_Def[3].Right);
    SetInputKey("P1C.Start","none",Keys_Def[3].Start);
    SetInputKey("P1C.A","none",Keys_Def[3].A);
    SetInputKey("P1C.B","none",Keys_Def[3].B);
    SetInputKey("P1C.C","none",Keys_Def[3].C);
    SetInputKey("P1C.Mode","none",Keys_Def[3].Mode);
    SetInputKey("P1C.X","none",Keys_Def[3].X);
    SetInputKey("P1C.Y","none",Keys_Def[3].Y);
    SetInputKey("P1C.Z","none",Keys_Def[3].Z); 

	Controller_1D_Type = GetPrivateProfileInt("Input", "P1D.Type", 0, Conf_File);
    SetInputKey("P1D.Up","none",Keys_Def[4].Up);
    SetInputKey("P1D.Down","none",Keys_Def[4].Down);
    SetInputKey("P1D.Left","none",Keys_Def[4].Left);
    SetInputKey("P1D.Right","none",Keys_Def[4].Right);
    SetInputKey("P1D.Start","none",Keys_Def[4].Start);
    SetInputKey("P1D.A","none",Keys_Def[4].A);
    SetInputKey("P1D.B","none",Keys_Def[4].B);
    SetInputKey("P1D.C","none",Keys_Def[4].C);
    SetInputKey("P1D.Mode","none",Keys_Def[4].Mode);
    SetInputKey("P1D.X","none",Keys_Def[4].X);
    SetInputKey("P1D.Y","none",Keys_Def[4].Y);
    SetInputKey("P1D.Z","none",Keys_Def[4].Z); 

	Controller_2_Type = GetPrivateProfileInt("Input", "P2.Type", 0, Conf_File);
    SetInputKey("P2.Up","none",Keys_Def[1].Up);
    SetInputKey("P2.Down","none",Keys_Def[1].Down);
    SetInputKey("P2.Left","none",Keys_Def[1].Left);
    SetInputKey("P2.Right","none",Keys_Def[1].Right);
    SetInputKey("P2.Start","none",Keys_Def[1].Start);
    SetInputKey("P2.A","none",Keys_Def[1].A);
    SetInputKey("P2.B","none",Keys_Def[1].B);
    SetInputKey("P2.C","none",Keys_Def[1].C);
    SetInputKey("P2.Mode","none",Keys_Def[1].Mode);
    SetInputKey("P2.X","none",Keys_Def[1].X);
    SetInputKey("P2.Y","none",Keys_Def[1].Y);
    SetInputKey("P2.Z","none",Keys_Def[1].Z); 

	Controller_2B_Type = GetPrivateProfileInt("Input", "P2B.Type", 0, Conf_File);
    SetInputKey("P2B.Up","none",Keys_Def[5].Up);
    SetInputKey("P2B.Down","none",Keys_Def[5].Down);
    SetInputKey("P2B.Left","none",Keys_Def[5].Left);
    SetInputKey("P2B.Right","none",Keys_Def[5].Right);
    SetInputKey("P2B.Start","none",Keys_Def[5].Start);
    SetInputKey("P2B.A","none",Keys_Def[5].A);
    SetInputKey("P2B.B","none",Keys_Def[5].B);
    SetInputKey("P2B.C","none",Keys_Def[5].C);
    SetInputKey("P2B.Mode","none",Keys_Def[5].Mode);
    SetInputKey("P2B.X","none",Keys_Def[5].X);
    SetInputKey("P2B.Y","none",Keys_Def[5].Y);
    SetInputKey("P2B.Z","none",Keys_Def[5].Z); 

	Controller_2C_Type = GetPrivateProfileInt("Input", "P2C.Type", 0, Conf_File);
    SetInputKey("P2C.Up","none",Keys_Def[6].Up);
    SetInputKey("P2C.Down","none",Keys_Def[6].Down);
    SetInputKey("P2C.Left","none",Keys_Def[6].Left);
    SetInputKey("P2C.Right","none",Keys_Def[6].Right);
    SetInputKey("P2C.Start","none",Keys_Def[6].Start);
    SetInputKey("P2C.A","none",Keys_Def[6].A);
    SetInputKey("P2C.B","none",Keys_Def[6].B);
    SetInputKey("P2C.C","none",Keys_Def[6].C);
    SetInputKey("P2C.Mode","none",Keys_Def[6].Mode);
    SetInputKey("P2C.X","none",Keys_Def[6].X);
    SetInputKey("P2C.Y","none",Keys_Def[6].Y);
    SetInputKey("P2C.Z","none",Keys_Def[6].Z); 

	Controller_2D_Type = GetPrivateProfileInt("Input", "P2D.Type", 0, Conf_File);
    SetInputKey("P2D.Up","none",Keys_Def[7].Up);
    SetInputKey("P2D.Down","none",Keys_Def[7].Down);
    SetInputKey("P2D.Left","none",Keys_Def[7].Left);
    SetInputKey("P2D.Right","none",Keys_Def[7].Right);
    SetInputKey("P2D.Start","none",Keys_Def[7].Start);
    SetInputKey("P2D.A","none",Keys_Def[7].A);
    SetInputKey("P2D.B","none",Keys_Def[7].B);
    SetInputKey("P2D.C","none",Keys_Def[7].C);
    SetInputKey("P2D.Mode","none",Keys_Def[7].Mode);
    SetInputKey("P2D.X","none",Keys_Def[7].X);
    SetInputKey("P2D.Y","none",Keys_Def[7].Y);
    SetInputKey("P2D.Z","none",Keys_Def[7].Z); 

	Make_IO_Table();
    
    //Modif IM - make ui buttons configurable
    SetInterfaceKey("Pause Emulator","escape",Pause_Emulator_key);
	SetInterfaceKey("Quit","s-escape",Quit_key);
	SetInterfaceKey("Take Screenshot","backspace",Take_Screenshot_key);
	SetInterfaceKey("Fast Forward","tab",Fast_Forward_key);
	SetInterfaceKey("Reset","s-tab",Reset_key);
	SetInterfaceKey("Lower Speed","[",Lower_Speed_key);
	SetInterfaceKey("Raise Speed","]",Raise_Speed_key);
	SetInterfaceKey("Toggle Fullscreen","a-return",Toggle_Fullscreen_key);
	SetInterfaceKey("Change Stretch","s-f2",Change_Stretch_key);
	SetInterfaceKey("Auto Frameskip","f2",Auto_Frameskip_key);
	SetInterfaceKey("Change Vsync","s-f3",Change_Vsync_key);
	SetInterfaceKey("Lower Frameskip","f3",Lower_Frameskip_key);
	SetInterfaceKey("Raise Frameskip","f4",Raise_Frameskip_key);
	SetInterfaceKey("Save Current State","f5",Save_Current_State_key);
	SetInterfaceKey("Alternate Save Current State","unknown",Alt_Save_Current_State_key);
	SetInterfaceKey("Select Previous State","f6",Select_Previous_State_key);
	SetInterfaceKey("Select Next State","f7",Select_Next_State_key);
	SetInterfaceKey("Load Current State","f8",Load_Current_State_key);
	SetInterfaceKey("Alternate Load Current State","unknown",Alt_Load_Current_State_key);
	SetInterfaceKey("Change Blit Style","s-f9",Change_Blit_Style_key);
	SetInterfaceKey("Toggle Fast Blur","f9",Toggle_Fast_Blur_key);
	SetInterfaceKey("Toggle Improved DAC","s-f10",Toggle_Improved_DAC_key);
	SetInterfaceKey("Toggle FPS Display","f10",Toggle_FPS_Display_key);
	SetInterfaceKey("Toggle Improved PSG","s-f11",Toggle_Improved_PSG_key);
	SetInterfaceKey("Previous Rendering Mode","f11",Previous_Rendering_Mode_key);
	SetInterfaceKey("Toggle Improved YM2612","s-f12",Toggle_Improved_YM2612_key);
	SetInterfaceKey("Next Rendering Mode","f12",Next_Rendering_Mode_key);
	SetInterfaceKey("Select State 0","c-0",Select_State_0_key);
	SetInterfaceKey("Select State 1","c-1",Select_State_1_key);
	SetInterfaceKey("Select State 2","c-2",Select_State_2_key);
	SetInterfaceKey("Select State 3","c-3",Select_State_3_key);
	SetInterfaceKey("Select State 4","c-4",Select_State_4_key);
	SetInterfaceKey("Select State 5","c-5",Select_State_5_key);
	SetInterfaceKey("Select State 6","c-6",Select_State_6_key);
	SetInterfaceKey("Select State 7","c-7",Select_State_7_key);
	SetInterfaceKey("Select State 8","c-8",Select_State_8_key);
	SetInterfaceKey("Select State 9","c-9",Select_State_9_key);
	SetInterfaceKey("Load State 0","0",Load_State_0_key);
	SetInterfaceKey("Load State 1","1",Load_State_1_key);
	SetInterfaceKey("Load State 2","2",Load_State_2_key);
	SetInterfaceKey("Load State 3","3",Load_State_3_key);
	SetInterfaceKey("Load State 4","4",Load_State_4_key);
	SetInterfaceKey("Load State 5","5",Load_State_5_key);
	SetInterfaceKey("Load State 6","6",Load_State_6_key);
	SetInterfaceKey("Load State 7","7",Load_State_7_key);
	SetInterfaceKey("Load State 8","8",Load_State_8_key);
	SetInterfaceKey("Load State 9","9",Load_State_9_key);
	SetInterfaceKey("Save State 0","s-0",Save_State_0_key);
	SetInterfaceKey("Save State 1","s-1",Save_State_1_key);
	SetInterfaceKey("Save State 2","s-2",Save_State_2_key);
	SetInterfaceKey("Save State 3","s-3",Save_State_3_key);
	SetInterfaceKey("Save State 4","s-4",Save_State_4_key);
	SetInterfaceKey("Save State 5","s-5",Save_State_5_key);
	SetInterfaceKey("Save State 6","s-6",Save_State_6_key);
	SetInterfaceKey("Save State 7","s-7",Save_State_7_key);
	SetInterfaceKey("Save State 8","s-8",Save_State_8_key);
	SetInterfaceKey("Save State 9","s-9",Save_State_9_key);
	SetInterfaceKey("Close ROM","c-c",Close_ROM_key);
	SetInterfaceKey("Open Game Genie","c-g",Open_Game_Genie_key);
	SetInterfaceKey("Change SegaCD Syncro","c-e",Change_SegaCD_Syncro_key);
	SetInterfaceKey("Start Playing GYM","cs-g",Start_Playing_GYM_key);
	SetInterfaceKey("Change CD","c-v",Change_CD_key);
	SetInterfaceKey("Toggle WAV Dumping","c-w",Toggle_WAV_Dumping_key);
	SetInterfaceKey("RAM Search Menu","a-r",RAM_Search_Menu_key);
	SetInterfaceKey("Start Recording","cs-r",Start_Recording_key);
	SetInterfaceKey("Toggle Readonly","r",Toggle_Readonly_key);
	SetInterfaceKey("Play Movie","p",Play_Movie_key);
	SetInterfaceKey("Toggle Frame Counter","f",Toggle_Frame_Counter_key);
	SetInterfaceKey("Advance Frame","i",Advance_Frame_key);
	SetInterfaceKey("Toggle Input Display","o",Toggle_Input_Display_key);
	SetInterfaceKey("Toggle Sound","s",Toggle_Sound_key);
	SetInterfaceKey("Stop Movie","q",Stop_Movie_key);
	SetInterfaceKey("Record New Movie","n",Record_New_Movie_key);
	SetInterfaceKey("Load State File","l",Load_State_File_key);
	SetInterfaceKey("Save State File","s-s",Save_State_File_key);
	SetInterfaceKey("Save Config File","c-s",Save_Config_File_key);

#ifndef __PORT__
	DestroyMenu(Gens_Menu);

	if (Full_Screen) Build_Main_Menu();
	else SetMenu(HWnd, Build_Main_Menu());		// Update new menu
#endif
	return 1;
}

//Modif IM - convert a human-readable key string to a SDL_keysym
void ascii2SDLK(char* str, SDL_keysym* ks)
{
    int mod = KMOD_NONE;
    int sym = 0;
    if (!strcmp(str,"none"))
    {
        ks->mod = (SDLMod)mod;
        ks->sym = (SDLKey)SDLK_NONE; //won't be triggered or confused for an unknown key
        return;
    }
    char* str0 = str;
    //if has a modifier (contains '-' after the first char but isn't "[-]")
    if (str[0] != '[' && index(str,'-') > str)
    {
        //translate the mod
        while(str[0] != '-')
        {
            if (str[0] == 'c')
                mod |= KMOD_CTRL;
            if (str[0] == 's')
                mod |= KMOD_SHIFT;
            if (str[0] == 'a')
                mod |= KMOD_ALT;
            if (str[0] == 'm')
                mod |= KMOD_META;
            str++;
        }
        str++;
    }

    //It's inefficient, but it's simple and will only run once at startup anyway.
    char* tmp;
    int max_len;
    for (int k=0; k < SDLK_LAST; k++)
    {
        tmp = SDL_GetKeyName((SDLKey)k);
        max_len = strlen(tmp)>strlen(str) ? strlen(tmp) : strlen(str);
        
        if ( !(strncmp(str,tmp,max_len) || sym))
        {
            sym = k;
            //printf("identified %s\n",tmp);
        }
    }

    if (sym == 0)
        printf("unknown key: %s\n",str0);

    ks->mod = (SDLMod)mod;
    ks->sym = (SDLKey)sym;
}

//opposite of above function            
void SDLK2ascii (char* str, SDL_keysym* ks)
{
    if (ks->sym == SDLK_NONE)
    {
        sprintf(str,"none");
        return;
    }
    //longest possible string is "csam-print screen"
    if( ks->mod & KMOD_CTRL)
        *str++ = 'c';
    if( ks->mod & KMOD_SHIFT)
        *str++ = 's';
    if( ks->mod & KMOD_ALT) 
        *str++ = 'a';
    if( ks->mod & KMOD_META) 
        *str++ = 'm';
    if( ks->mod & 0xE000) //These shouldn't appear.
        *str++ = 'X';
    if (ks->mod & (SDLMod)(KMOD_CTRL|KMOD_SHIFT|KMOD_ALT|KMOD_META))
        *str++ = '-';
    sprintf(str,"%s",SDL_GetKeyName(ks->sym));
}

int keysyms_match(SDL_keysym* k1, SDL_keysym* k2)
{
    //'!' is so I compare whether they're true/false
    if ((k1->sym == k2->sym) &&
        ( !(k1->mod & KMOD_CTRL  ) == !(k2->mod & KMOD_CTRL  )) &&
        ( !(k1->mod & KMOD_ALT   ) == !(k2->mod & KMOD_ALT   )) &&
        ( !(k1->mod & KMOD_SHIFT ) == !(k2->mod & KMOD_SHIFT )) &&
        ( !(k1->mod & KMOD_META  ) == !(k2->mod & KMOD_META  )) )
        return 1;
    return 0;
}

void print_SDLK(SDL_keysym* k)
{
    printf("%s: ",SDL_GetKeyName(k->sym));
    if( k->mod & KMOD_NUM ) printf( "NUMLOCK " );
    if( k->mod & KMOD_CAPS ) printf( "CAPSLOCK " );
    if( k->mod & KMOD_LCTRL ) printf( "LCTRL " );
    if( k->mod & KMOD_RCTRL ) printf( "RCTRL " );
    if( k->mod & KMOD_RSHIFT ) printf( "RSHIFT " );
    if( k->mod & KMOD_LSHIFT ) printf( "LSHIFT " );
    if( k->mod & KMOD_RALT ) printf( "RALT " );
    if( k->mod & KMOD_LALT ) printf( "LALT " );
    if( k->mod & KMOD_CTRL ) printf( "CTRL " );
    if( k->mod & KMOD_SHIFT ) printf( "SHIFT " );
    if( k->mod & KMOD_ALT ) printf( "ALT " );
    if( ! k->mod ) printf( "NONE " );
    printf( "\n" );
}

int Load_As_Config(HWND hWnd, void *Game_Active)
{
	char Name[2048];

#ifdef __PORT__
#ifdef WITH_GTK
	
	GtkWidget* filesel;
	gint res;

	filesel = create_file_selector_private(NULL/*Filter*/,0);	
	res = gtk_dialog_run(GTK_DIALOG(filesel));
	
	switch(res)
	{
		case GTK_RESPONSE_OK:
			strcpy(Name,gtk_file_selection_get_filename(GTK_FILE_SELECTION(filesel)));
			break;
		case GTK_RESPONSE_CANCEL:
			break;
	}

	gtk_widget_destroy(filesel);
	
#endif	
#else
	char Name[2048];
	OPENFILENAME ofn;

	SetCurrentDirectory(Gens_Path);

	strcpy(&Name[0], "Gens.cfg");
	memset(&ofn, 0, sizeof(OPENFILENAME));

	ofn.lStructSize = sizeof(OPENFILENAME);
	ofn.hwndOwner = hWnd;
	ofn.hInstance = ghInstance;
	ofn.lpstrFile = Name;
	ofn.nMaxFile = 2047;
	ofn.lpstrFilter = "Config Files\0*.cfg\0All Files\0*.*\0\0";
	ofn.nFilterIndex = 1;
	ofn.lpstrInitialDir = Gens_Path;
	ofn.lpstrDefExt = "cfg";
	ofn.Flags = OFN_FILEMUSTEXIST;

#endif
#ifdef __PORT__
#ifdef WITH_GTK
	if (strlen(Name))
#endif
#else
	if(GetOpenFileName(&ofn))
#endif
	{
		Load_Config(Name, Game_Active);
		strcpy(Str_Tmp, "config loaded from ");
		strcat(Str_Tmp, Name);
		Put_Info(Str_Tmp, 2000);
		return 1;
	}

	return 0;
}


int Load_SRAM(void)
{
	FILE *SRAM_File = 0;
	char Name[2048];

	memset(SRAM, 0, 64 * 1024);

	strcpy(Name, SRAM_Dir);
	strcat(Name, Rom_Name);
	strcat(Name, ".srm");
	
	if((SRAM_File = fopen(Name, "rb")) == 0) return 0;
	fread(SRAM, 64 * 1024, 1, SRAM_File);
	fclose(SRAM_File);

	strcpy(Str_Tmp, "SRAM loaded from ");
	strcat(Str_Tmp, Name);
	Put_Info(Str_Tmp, 2000);
	return 1;
}

int Save_SRAM(void)
{
	FILE *SRAM_File = 0;
	int size_to_save, i;
	char Name[2048];

	i = (64 * 1024) - 1;
	while ((i >= 0) && (SRAM[i] == 0)) i--;

	if (i < 0) return 0;

	i++;

	size_to_save = 1;
	while (i > size_to_save) size_to_save <<= 1;

	strcpy(Name, SRAM_Dir);
	strcat(Name, Rom_Name);
	strcat(Name, ".srm");
	
	if((SRAM_File = fopen(Name, "wb")) == 0) return 0;
	
	fwrite(SRAM, size_to_save, 1, SRAM_File);
	fclose(SRAM_File);

	strcpy(Str_Tmp, "SRAM saved in ");
	strcat(Str_Tmp, Name);
	Put_Info(Str_Tmp, 2000);

	return 1;
}


void S_Format_BRAM(unsigned char *buf)
{
	memset(buf, 0x5F, 11);

	buf[0x0F] = 0x40;

	buf[0x11] = 0x7D;
	buf[0x13] = 0x7D;
	buf[0x15] = 0x7D;
	buf[0x17] = 0x7D;

	sprintf((char *) &buf[0x20], "SEGA CD ROM");
	sprintf((char *) &buf[0x30], "RAM CARTRIDGE");

	buf[0x24] = 0x5F;
	buf[0x27] = 0x5F;

	buf[0x2C] = 0x01;

	buf[0x33] = 0x5F;
	buf[0x3D] = 0x5F;
	buf[0x3E] = 0x5F;
	buf[0x3F] = 0x5F;
}


void Format_Backup_Ram(void)
{
	memset(Ram_Backup, 0, 8 * 1024);

	S_Format_BRAM(&Ram_Backup[0x1FC0]);

	memset(Ram_Backup_Ex, 0, 64 * 1024);
}


int Load_BRAM(void)
{
	FILE *BRAM_File = 0;
	char Name[2048];

	Format_Backup_Ram();

	strcpy(Name, BRAM_Dir);
	strcat(Name, Rom_Name);
	strcat(Name, ".brm");
	
	if((BRAM_File = fopen(Name, "rb")) == 0) return 0;
	
	fread(Ram_Backup, 8 * 1024, 1, BRAM_File);
	fread(Ram_Backup_Ex, (8 << BRAM_Ex_Size) * 1024, 1, BRAM_File);
	fclose(BRAM_File);

	strcpy(Str_Tmp, "BRAM loaded from ");
	strcat(Str_Tmp, Name);
	Put_Info(Str_Tmp, 2000);
	return 1;
}

int Save_BRAM(void)
{
	FILE *BRAM_File = 0;
	char Name[2048];

	strcpy(Name, BRAM_Dir);
	strcat(Name, Rom_Name);
	strcat(Name, ".brm");
	
	if((BRAM_File = fopen(Name, "wb")) == 0) return 0;
	
	fwrite(Ram_Backup, 8 * 1024, 1, BRAM_File);
	fwrite(Ram_Backup_Ex, (8 << BRAM_Ex_Size) * 1024, 1, BRAM_File);
	fclose(BRAM_File);

	strcpy(Str_Tmp, "BRAM saved in ");
	strcat(Str_Tmp, Name);
	Put_Info(Str_Tmp, 2000);
	return 1;
}
